import datetime
import json
import math
import os
import tempfile
import time
from collections import deque

import cv2
import numpy as np
import torch
import torch.nn as nn
from pycocotools.cocoeval import COCOeval
from torch.cuda.amp import GradScaler, autocast
from tqdm import tqdm

from tools.GLOBAL_LOG import flog
from tools.boxes.f_boxes import ltrb2ltwh
from tools.fits.fcocoeval import FCOCOeval
from tools.picture.f_show import f_show_od_np4plt_v3


class FModelBase(nn.Module):
    def __init__(self, net, losser, preder):
        super(FModelBase, self).__init__()
        self.net = net
        self.losser = losser
        self.preder = preder

    def forward(self, datas_batch):
        imgs_ts_4d, targets = datas_batch
        outs = self.net(imgs_ts_4d)

        ''' 多尺度图片归一化支持 '''
        device = imgs_ts_4d.device
        batch, c, h, w = imgs_ts_4d.shape
        toones_wh_ts_input = []
        if targets is not None and 'toone' in targets[0]:  # 多尺度归一
            for target in targets:
                toones_wh_ts_input.append(torch.tensor(target['toone'], device=device))
            toones_wh_ts_input = torch.stack(toones_wh_ts_input, 0)
        else:
            toones_wh_ts_input = torch.empty(batch, 2, device=device, dtype=torch.float)  # 实际为准
            toones_wh_ts_input[:, :] = torch.tensor(imgs_ts_4d.shape[2:4][::-1], device=device)

        if self.training:
            if targets is None:
                raise ValueError("In training mode, targets should be passed")
            loss_total, log_dict = self.losser(outs, targets, imgs_ts_4d, toones_wh_ts_input)
            '''------ 验证 loss 由FitExecutor 处理 ------'''

            return loss_total, log_dict
        else:
            with torch.no_grad():  # 这个没用
                # outs模型输出  返回 ids_batch, p_boxes_ltrb, p_keypoints, p_labels, p_scores
                reses = self.preder(outs, imgs_ts_4d, targets, toones_wh_ts_input)

            return reses, imgs_ts_4d, targets, toones_wh_ts_input


class FitExecutor():
    def __init__(self, cfg, model, fun_loss, save_weight_name, path_save_weight,
                 optimizer=None, dataloader_train=None, lr_scheduler=None,
                 end_epoch=None, is_mixture_fit=True, lr_val_base=1e-3,
                 is_writer=False,
                 num_save_interval=10, print_freq=10,
                 dataloader_val=None,
                 dataloader_test=None, maps_def_max=None,
                 num_vis_z=None, mode_vis=None,
                 ) -> None:
        '''

        :param model:
        :param fun_loss: 回调 loss 处理
        :param save_weight_name:
        :param path_save_weight:
        :param optimizer:
        :param dataloader_train:
        :param lr_scheduler:
        :param end_epoch:
        :param is_mixture_fit:
        :param lr_val_base:
        :param is_writer:
        :param num_save_interval:
        :param print_freq:
        :param dataloader_val:
        :param dataloader_test:
        '''
        super().__init__()
        self.dataloader_train = dataloader_train
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.end_epoch = end_epoch
        self.is_mixture_fit = is_mixture_fit
        self.lr_val_base = lr_val_base

        self.fun_loss = fun_loss
        self.cfg = cfg

        # 测试属性
        self.dataloader_test = dataloader_test
        self.num_vis_z = num_vis_z
        self.mode_vis = mode_vis  # 'keypoints','bbox'
        self.maps_def_max = maps_def_max

        # 验证
        self.dataloader_val = dataloader_val

        assert os.path.exists(path_save_weight), 'path_save_weight 不存在 %s' % path_save_weight
        self.save_weight_name = save_weight_name
        self.path_save_weight = path_save_weight
        self.num_save_interval = num_save_interval
        self.print_freq = print_freq

        if is_writer:

            from torch.utils.tensorboard import SummaryWriter
            c_time = time.strftime('%Y-%m-%d_%H_%M_%S', time.localtime(time.time()))
            _path = os.path.join(path_save_weight, c_time)

            flog.debug('---- use tensorboard --- %s', _path)
            os.makedirs(_path, exist_ok=True)
            self.tb_writer = SummaryWriter(_path)
        else:
            self.tb_writer = None

    def frun(self, start_epoch):
        for epoch in range(start_epoch, self.end_epoch + 1, 1):  # 从1开始
            save_val = None
            ''' ------------------- 训练代码  --------------------- '''
            if self.dataloader_train is not None and self.cfg.IS_TRAIN:
                t0_test = time.time()
                loss_val_obj = self.ftrain(dataloader_train=self.dataloader_train,
                                           optimizer=self.optimizer,
                                           epoch=epoch, end_epoch=self.end_epoch,
                                           fun_loss=self.fun_loss, model=self.model,
                                           lr_val_base=self.lr_val_base, tb_writer=self.tb_writer,
                                           is_mixture_fit=self.is_mixture_fit, print_freq=self.print_freq)
                save_val = loss_val_obj.avg

                print(' ----- 训练用时 %s ----- ' % str(datetime.timedelta(seconds=int(time.time() - t0_test))))

            if save_val is not None \
                    and self.dataloader_train is not None \
                    and (epoch % self.num_save_interval) == 0:
                print('训练完成正在保存模型...')
                save_weight(
                    path_save=self.path_save_weight,
                    model=self.model,
                    name=self.save_weight_name,
                    loss=save_val,  # 这个正常是一样的有的
                    optimizer=self.optimizer,
                    lr_scheduler=self.lr_scheduler,
                    epoch=epoch)

            ''' ------------------- 验证代码  --------------------- '''
            if self.dataloader_val is not None and self.cfg.IS_VAL and self.is_open(epoch, self.cfg.NUMS_VAL_DICT):
                t0_test = time.time()
                loss_val_obj = self.fval(dataloader_val=self.dataloader_val,
                                         epoch=epoch, end_epoch=self.end_epoch,
                                         fun_loss=self.fun_loss, model=self.model,
                                         tb_writer=self.tb_writer, print_freq=self.print_freq
                                         )
                print(' ----- 验证用时 %s ----- ' % str(datetime.timedelta(seconds=int(time.time() - t0_test))))

            ''' ------------------- 测试代码  --------------------- '''
            if self.dataloader_test is not None and self.cfg.IS_TEST and self.is_open(epoch, self.cfg.NUMS_TEST_DICT):
                t0_test = time.time()
                maps_val = self.ftest(model=self.model, dataloader_test=self.dataloader_test,
                                      epoch=epoch, is_vis_all=self.cfg.IS_VISUAL, mode_vis=self.mode_vis,
                                      num_vis_z=self.num_vis_z, tb_writer=None, )

                if maps_val is not None and self.maps_def_max is not None:
                    # 更新 self.maps_def_max 值
                    if maps_val[0] > self.maps_def_max[0]:
                        self.maps_def_max[0] = maps_val[0]
                        self.maps_def_max[1] = max(self.maps_def_max[1], maps_val[1])
                    elif maps_val[1] > self.maps_def_max[1]:
                        self.maps_def_max[0] = max(self.maps_def_max[0], maps_val[0])
                        self.maps_def_max[1] = maps_val[1]
                    else:
                        continue
                    save_weight(
                        path_save=self.path_save_weight,
                        model=self.model,
                        name=self.save_weight_name,
                        loss=save_val,  # 这个正常是一样的有的
                        optimizer=self.optimizer,
                        lr_scheduler=self.lr_scheduler,
                        epoch=epoch,
                        maps_val=self.maps_def_max,
                    )
                print(' ----- 测试用时 %s ----- ' % str(datetime.timedelta(seconds=int(time.time() - t0_test))))

    def fval(self, dataloader_val, epoch, end_epoch,
             model=None, fun_loss=None,
             print_freq=1, tb_writer=None,
             ):
        # 这个在 epoch 中
        loss_val_obj = SmoothedValue()
        print('\n-------------------- 验证 fval 开始 %s -------------------------' % epoch)
        epoch_size = len(dataloader_val)
        batch = dataloader_val.batch_size

        t0 = time.time()
        for i, datas_batch in enumerate(dataloader_val):
            t1 = time.time()

            with torch.no_grad():
                if fun_loss is not None:
                    # 这里是回调
                    loss_total, log_dict = fun_loss(datas_batch)
                else:
                    loss_total, log_dict = model(datas_batch)

            loss_val_obj.update(loss_total.item())

            if i % (print_freq * batch) == 0:
                self.print_log(end_epoch=end_epoch, epoch=epoch,
                               epoch_size=epoch_size, iter_i=i,
                               log_dict=log_dict, loss_val=loss_val_obj.avg,
                               t0=t0, t1=t1, title='val')

            if tb_writer is not None:
                self.tb_writing4train_val(tb_writer=tb_writer, log_dict=log_dict,
                                          epoch=epoch, epoch_size=epoch_size, iter_i=i, lr=None)

            # 更新时间用于获取 data 时间
            t0 = time.time()

        return loss_val_obj

    def ftrain(self, dataloader_train, optimizer, epoch, end_epoch,
               fun_loss=None, model=None,
               lr_val_base=1e-3, tb_writer=None,
               is_mixture_fit=False, print_freq=1):
        # 这个在 epoch 中
        loss_val_obj = SmoothedValue()
        print('-------------------- 训练 ftrain 开始 %s -------------------------' % epoch)
        epoch_size = len(dataloader_train)
        batch = dataloader_train.batch_size
        scaler = GradScaler(enabled=is_mixture_fit)

        t0 = time.time()
        for i, datas_batch in enumerate(dataloader_train):
            t1 = time.time()

            if epoch < 2:
                now_lr = lr_val_base * pow((i + epoch * epoch_size) * 1. / (1 * epoch_size), 4)
                self.update_lr(optimizer, now_lr)
            elif epoch == 1:
                self.update_lr(optimizer, lr_val_base)

            with autocast(enabled=is_mixture_fit):
                if fun_loss is not None:
                    # 这里是回调
                    loss_total, log_dict = fun_loss(datas_batch)
                else:
                    loss_total, log_dict = model(datas_batch)

            loss_val_obj.update(loss_total.item())

            scaler.scale(loss_total).backward()
            scaler.step(optimizer)
            scaler.update()
            optimizer.zero_grad()

            if i % (print_freq * batch) == 0:
                self.print_log(end_epoch=end_epoch, epoch=epoch,
                               epoch_size=epoch_size, iter_i=i,
                               log_dict=log_dict, loss_val=loss_val_obj.avg,
                               lr=optimizer.param_groups[0]['lr'],
                               t0=t0, t1=t1, title='train')

            if tb_writer is not None:
                self.tb_writing4train_val(tb_writer=tb_writer, log_dict=log_dict,
                                          epoch=epoch, epoch_size=epoch_size, iter_i=i,
                                          lr=optimizer.param_groups[0]['lr'])

            # 更新时间用于获取 data 时间
            t0 = time.time()

        return loss_val_obj

    def ftest(self, model, dataloader_test, epoch, is_vis_all, mode_vis, num_vis_z, tb_writer=None):
        '''
        ['segm', 'bbox', 'keypoints']
        :param epoch:
        :return:
        '''
        print('-------------------- 测试 ftest 开始 %s -------------------------' % epoch)
        model.eval()

        num_no_pos_z = 0  # 没发现目标的计数器
        res_z = {}  # 保存 全部的coco结果
        num_vis_now = 0  # 已可视化的临时值
        ids_data_all = []  # dataloader_test 的所有图片ID

        # with torch.no_grad() 这个由模型处理
        pbar = tqdm(dataloader_test, desc='%s' % epoch, postfix=dict, mininterval=0.1)
        for datas_batch in pbar:
            # 由模型 FModelBase 处理数据后输出
            reses, imgs_ts_4d, targets, toones_wh_ts_input = model(datas_batch)
            # 这里输出的都是归一化尺寸
            ids_batch, p_ltrbs, p_kps, p_labels, p_scores = reses

            ''' 整批没有目标 提示和处理 '''
            if p_labels is None or len(p_labels) == 0:
                num_no_pos_z += len(dataloader_test)
                flog.info('本批没有目标 当前共有: %s 张图片未检出', num_no_pos_z)
                # if num_no_pos > 3:  # 如果3个批都没有目标则放弃
                #     return
                # else:  # 没有目标就下一个
                #     num_no_pos += 1
                pbar.set_description("未检出数: %s" % num_no_pos_z)
                continue

            ''' 有目标进行如下操作 提取该批真实ID及尺寸 '''
            size_f_batch = []  # 用于修复box
            ids_img_batch = []
            for target in targets:  # 通过target 提取ID 和 size
                ids_data_all.append(target['image_id'])
                ids_img_batch.append(target['image_id'])
                size_f_batch.append(target['size'].clone().detach())  # tnesor

            _res_batch = {}  # 每一批的coco 结果 临时值

            # 每一张图的 id 与批次顺序保持一致 选出匹配
            for i, (size, image_id) in enumerate(zip(size_f_batch, ids_img_batch)):
                # 图片与 输出的对应
                mask = ids_batch == i  # 构建 batch 次的mask
                ''' 单张图没有目标 计数提示 '''
                if not torch.any(mask):
                    # flog.warning('没有预测出框 %s', files_txt)
                    num_no_pos_z += 1
                    pbar.set_description("未检出数: %s" % num_no_pos_z)
                else:
                    ''' 已检出 是否可视化逻辑 '''
                    if is_vis_all or num_vis_now < num_vis_z:
                        num_vis_now += 1
                        # 多尺度处理逻辑
                        if 'toone' in targets[i]:
                            size_wh_toone_ts_x2 = targets[i]['toone'].repeat(2)  # input尺寸
                        else:
                            size_wh_toone_ts_x2 = torch.tensor(imgs_ts_4d.shape[-2:][::-1]).repeat(2)  # 实际为准
                        self.show_pic(dataloader_test=dataloader_test,
                                      size_wh_toone_ts_x2=size_wh_toone_ts_x2,
                                      target=targets[i], image_id=image_id, mode_vis=mode_vis,
                                      p_labels_pos=p_labels[mask],
                                      p_ltrbs_pos=p_ltrbs[mask], p_scores_pos=p_scores[mask])

                    # 归一化->file尺寸  coco需要 ltwh
                    boxes_ltwh = ltrb2ltwh(p_ltrbs[mask] * size.repeat(2)[None])
                    _res_batch[image_id] = {
                        'boxes': boxes_ltwh,  # coco loadRes 会对ltwh 转换成 ltrb
                        'labels': p_labels[mask],
                        'scores': p_scores[mask],
                    }
                    # 更新进度条值
                    d = {
                        'pos': len(boxes_ltwh),  # 框个数
                        'max': round(p_scores[mask].max().item() * 100, 1),  # 分数
                        # 'min': round(p_scores.min().item(), 1),
                        'mean': round(p_scores[mask].mean().item() * 100, 1),
                    }
                    pbar.set_postfix(**d)

            if len(_res_batch) > 0:  # 有检测出的目标
                res_z.update(_res_batch)  # 字典叠加

        res_coco_standard = []  # 最终的 coco 标准格式 一个ID可能 有多个目标
        # res_z 每一个ID可能有多个目标 每个目标形成一条 id对应数据
        for i, (image_id, g_target) in enumerate(res_z.items()):
            labels = g_target['labels'].type(torch.int).tolist()
            boxes_ltwh = g_target['boxes'].tolist()
            score = g_target['scores'].tolist()
            for i in range(len(labels)):
                # catid转换
                category_id = dataloader_test.dataset.classes_train2coco[labels[i]]
                res_coco_standard.append(
                    {"image_id": image_id, "category_id": category_id, "bbox": boxes_ltwh[i], "score": score[i]})

        if len(res_coco_standard) > 0:  # 有 coco 结果
            # 这个保存 AP50的结果
            maps_val = self.run_cocoeval(dataloader_test=dataloader_test, epoch=epoch,
                                         ids_data_all=ids_data_all,
                                         num_no_pos_z=num_no_pos_z,
                                         res_coco_standard=res_coco_standard,
                                         tb_writer=tb_writer)

        else:
            # 没有 coco 结果
            if tb_writer is not None:
                self.tr_writing4test(epoch=epoch, num_no_pos_z=num_no_pos_z,
                                     tb_writer=tb_writer, coco_eval_obj=None)
            maps_val = [0, 0]

        return maps_val

    def print_log(self, end_epoch, epoch,
                  epoch_size, iter_i, log_dict, loss_val,
                  t0, t1, lr=math.nan, title='title', ):
        '''

        :param end_epoch:
        :param epoch:
        :param epoch_size: 一个 epoch 需迭代的次数
        :param iter_i:  当前迭代次数
        :param log_dict:
        :param loss_val:
        :param lr: 当前学习率
        :param t0: 最开始的时间 这个是秒
        :param t1:  数据加载完成时间
        :param title:  标题
        :return:
        '''
        show_loss_str = []
        for k, v, in log_dict.items():
            show_loss_str.append(
                "{}: {:.4f} ||".format(k, v)
            )

        iter_time = time.time() - t0
        residue_time = iter_time * (epoch_size - iter_i + 1)  # 剩余时间

        d = {
            'title': title,
            'epoch': epoch,
            'end_epoch': end_epoch,
            'iter_i': iter_i + 1,
            'iter_epoch': epoch_size,
            'iter_z': (epoch - 1) * epoch_size + iter_i + 1,
            'lr': lr,
            'loss': loss_val,
            'loss_str': str(show_loss_str),
            'data_time': t1 - t0,
            'iter_time': iter_time,
            'residue_time': str(datetime.timedelta(seconds=int(residue_time))),
        }

        s = '[{title} {epoch}/{end_epoch}] ' \
            '[Iter {iter_i}/{iter_epoch}/{iter_z}] ' \
            '[lr: {lr:.6f}] [Loss: {loss:.2f} || {loss_str}] ' \
            '[time: {data_time:.2f}/{iter_time:.2f}/{residue_time}]'

        print(s.format(**d), flush=True)

    def update_lr(self, optimizer, now_lr):
        for param_group in optimizer.param_groups:
            param_group['lr'] = now_lr

    def is_open(self, epoch, nums_dict):
        '''
        是否开启 验证或测试
        :param epoch:
        :param nums_dict: NUMS_TEST_DICT = {2: 1, 35: 1, 50: 1}
        :return:
        '''

        #  {2: 1, 35: 1, 50: 1}  -> 2,35,50 ->  50,35,2
        s_keys = sorted(list(nums_dict), reverse=True)
        for s_key in s_keys:
            if epoch < s_key:
                continue
            else:
                eval_interval = nums_dict[s_key]
                if epoch % eval_interval != 0:
                    # 满足epoch 无需验证退出
                    break
                return True

    def run_cocoeval(self, dataloader_test, epoch, ids_data_all, num_no_pos_z, res_coco_standard, tb_writer):
        maps_val = []
        coco_gt = dataloader_test.dataset.coco_obj
        # 第一个元素指示操作该临时文件的安全级别，第二个元素指示该临时文件的路径
        _, tmp = tempfile.mkstemp()  # 创建临时文件
        json.dump(res_coco_standard, open(tmp, 'w'))
        coco_dt = coco_gt.loadRes(tmp)
        '''
                    _summarizeDets()->_summarize() 
                        _summarizeDets 函数中调用了12次 _summarize
                        结果在 self.eval['precision'] , self.eval['recall']中 
                    '''
        coco_eval_obj = FCOCOeval(coco_gt, coco_dt, 'bbox')  # 这个添加了每个类别的map分
        # coco_eval_obj = COCOeval(coco_gt, coco_dt, ann_type)
        coco_eval_obj.params.imgIds = ids_data_all  # 多显卡id合并更新
        coco_eval_obj.evaluate()
        coco_eval_obj.accumulate()
        coco_stats, print_coco = coco_eval_obj.summarize()
        print(print_coco)
        clses_name = list(dataloader_test.dataset.classes_ids)
        coco_eval_obj.print_clses(clses_name)
        maps_val.append(coco_eval_obj.stats[1])  # 添加ap50
        maps_val.append(coco_eval_obj.stats[7])
        if tb_writer is not None:
            # Precision_iou
            self.tr_writing4test(epoch=epoch, num_no_pos_z=num_no_pos_z,
                                 tb_writer=tb_writer, coco_eval_obj=coco_eval_obj)
        return maps_val

    def tr_writing4test(self, epoch, num_no_pos_z, tb_writer, coco_eval_obj=None):
        if coco_eval_obj is not None:
            _d = {
                'IoU=0.50:0.95': coco_eval_obj.stats[0],
                'IoU=0.50': coco_eval_obj.stats[1],
                'IoU=0.75': coco_eval_obj.stats[2],
            }
            tb_writer.add_scalars('mAP/Precision_iou', _d, epoch + 1)
            # Recall_iou
            _d = {
                'maxDets=  1': coco_eval_obj.stats[6],
                'maxDets= 10': coco_eval_obj.stats[7],
                'maxDets=100': coco_eval_obj.stats[8],
            }
            tb_writer.add_scalars('mAP/Recall_iou', _d, epoch + 1)
            # 小中大
            _d = {
                'p_large': coco_eval_obj.stats[5],
                'r_large': coco_eval_obj.stats[11],
            }
            tb_writer.add_scalars('mAP/large', _d, epoch + 1)
            _d = {
                'p_medium': coco_eval_obj.stats[4],
                'r_medium': coco_eval_obj.stats[10],
            }
            tb_writer.add_scalars('mAP/medium', _d, epoch + 1)
            _d = {
                'p_small': coco_eval_obj.stats[3],
                'r_small': coco_eval_obj.stats[9],
            }
            tb_writer.add_scalars('mAP/small', _d, epoch + 1)
        # 一个图只有一个值
        tb_writer.add_scalar('mAP/num_no_pos', num_no_pos_z, epoch + 1)  # 未检出的图片数

    def tb_writing4train_val(self, tb_writer, epoch, epoch_size, iter_i, log_dict, lr=None):
        # 主进程写入   不验证时写
        iter = epoch_size * (epoch - 1) + iter_i + 1
        for k, v, in log_dict.items():
            tb_writer.add_scalar('loss_iter/%s' % k, v, iter)
        if lr is not None:
            tb_writer.add_scalar('loss_iter/lr', lr, iter)

    def show_pic(self, dataloader_test, size_wh_toone_ts_x2,
                 target, image_id, mode_vis,
                 p_labels_pos, p_ltrbs_pos, p_scores_pos):

        coco = dataloader_test.dataset.coco_obj
        img_info = coco.loadImgs([image_id])
        file_img = os.path.join(dataloader_test.dataset.path_img, img_info[0]['file_name'])
        img_np_file = cv2.imread(file_img)
        img_np_file = cv2.cvtColor(img_np_file, cv2.COLOR_BGR2RGB)
        # import skimage.io as io
        # h,w,c
        # img_np = io.imread(file_img)
        size_wh_file = np.array(img_np_file.shape[:2][::-1])
        size_wh_file_x2 = np.tile(size_wh_file, 2)  # 图片真实尺寸
        p_boxes_ltrb_f = p_ltrbs_pos.cpu() * size_wh_file_x2

        g_texts = []
        for i, p_label in enumerate(p_labels_pos):
            name_cat = dataloader_test.dataset.ids_classes[(p_label.long() + 1).item()]
            s = name_cat + ':' + str(p_scores_pos[i].cpu().item())
            g_texts.append(s)

        if mode_vis == 'bbox':  # 单人脸
            f_show_od_np4plt_v3(
                img_np_file, p_ltrb=p_boxes_ltrb_f,
                g_ltrb=target['boxes'].cpu() / size_wh_toone_ts_x2 * size_wh_file_x2,
                g_texts=g_texts,
                is_recover_size=False,
            )
        elif mode_vis == 'keypoints':
            size_wh_file_x5 = np.tile(size_wh_file, self.cfg.NUM_KEYPOINTS)  # 图片真实尺寸
            raise Exception('待完善 keypoints')
            # p_keypoints_f = p_kps_pos.cpu() * size_wh_file_x5
            # f_show_kp_np4plt(img_np_file, p_boxes_ltrb_f,
            #                  kps_xy_input=p_keypoints_f,
            #                  mask_kps=torch.ones_like(p_keypoints_f, dtype=torch.bool),
            #                  # 测试集 GT默认不归一化,是input模型尺寸
            #                  g_ltrb=target['boxes'].cpu() / size_wh_toone_ts_x2 * size_wh_file_x2,
            #                  plabels_text=p_labels_pos,
            #                  p_scores_float=p_scores_pos.tolist(),
            #                  is_recover_size=False)
        else:
            raise Exception('self.cfg.MODE_VIS = %s 错误' % self.cfg.MODE_VIS)


class SmoothedValue(object):
    """
    记录一系列统计量
    Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{median:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)  # deque简单理解成加强版list
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    # def synchronize_between_processes(self):
    #     """
    #     Warning: does not synchronize the deque!
    #     """
    #     if not fis_mgpu():
    #         return
    #     t = torch.tensor([self.count, self.total], dtype=torch.float64, device="cuda")
    #     dist.barrier()
    #     dist.all_reduce(t)
    #     t = t.tolist()
    #     self.count = int(t[0])
    #     self.total = t[1]

    @property
    def median(self):  # @property 是装饰器，这里可简单理解为增加median属性(只读)
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value)


def save_weight(path_save, model, name, loss=None, optimizer=None, lr_scheduler=None, epoch=0, maps_val=None):
    if path_save and os.path.exists(path_save):
        sava_dict = {
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict() if optimizer else None,
            'lr_scheduler': lr_scheduler.state_dict() if lr_scheduler else None,
            'epoch': epoch}
        if maps_val is not None:
            if loss is not None:
                l = round(loss, 2)
            else:
                l = ''
            file_weight = os.path.join(path_save, (name + '-{}_{}_{}_{}.pth')
                                       .format(epoch + 1,
                                               l,
                                               'p' + str(round(maps_val[0] * 100, 1)),
                                               'r' + str(round(maps_val[1] * 100, 1)),
                                               ))
        else:
            file_weight = os.path.join(path_save, (name + '-{}_{}.pth').format(epoch, round(loss, 3)))
        torch.save(sava_dict, file_weight)
        flog.info('保存成功 %s', file_weight)


def load_weight(file_weight, model, optimizer=None, lr_scheduler=None,
                device=torch.device('cpu'), is_mgpu=False, ffun=None):
    start_epoch = 1
    if file_weight and os.path.exists(file_weight):
        checkpoint = torch.load(file_weight, map_location=device)

        '''对多gpu的k进行修复'''
        # if 'model' in checkpoint:
        #     pretrained_dict_y = checkpoint['model']
        # else:
        #     pretrained_dict_y = checkpoint

        ''' debug '''
        # if ffun is not None:
        #     pretrained_dict = ffun(pretrained_dict_y)
        # else:
        #     pretrained_dict = pretrained_dict_y

        dd = {}

        # # 多GPU处理
        # ss = 'module.'
        # for k, v in pretrained_dict.items():
        #     if is_mgpu:
        #         if ss not in k:
        #             dd[ss + k] = v
        #         else:
        #             dd = pretrained_dict_y
        #             break
        #             # dd[k] = v
        #     else:
        #         dd[k.replace(ss, '')] = v

        '''重组权重'''
        # load_weights_dict = {k: v for k, v in weights_dict.items()
        #                      if model.state_dict()[k].numel() == v.numel()}

        keys_missing, keys_unexpected = model.load_state_dict(dd, strict=False)
        if len(keys_missing) > 0 or len(keys_unexpected):
            flog.error('missing_keys %s', keys_missing)  # 这个是 model 的属性
            flog.error('unexpected_keys %s', keys_unexpected)  # 这个是 pth 的属性
        if optimizer:
            optimizer.load_state_dict(checkpoint['optimizer'])
        if lr_scheduler and checkpoint['lr_scheduler']:
            lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])

        if 'epoch' in checkpoint:
            start_epoch = checkpoint['epoch']
        else:
            start_epoch = 1

        flog.warning('已加载 feadre 权重文件为 %s', file_weight)
    return start_epoch



class FCFG_BASE:
    # 训练验证
    IS_MIXTURE_FIT = True  # 半精度训练
    IS_TRAIN = True  # 开启训练
    IS_VAL = True  # 开启验证

    IS_TEST = True  # 开启测试
    IS_VISUAL = False  # 可视化模式 关联训练 和测试
    MODE_VIS = 'bbox'  # 验证的 计算模式 'keypoints','bbox' 关系到 可视化 nms等
    NUM_VIS_Z = 3  # 测试时可视化的图片数 超过就不再可视

    IS_DEBUG = False  # 调试模式

    PTOPK = 500  # 预测时 取分数最高的前500个进行

    # 频率参数
    PRINT_FREQ = 1  # 打印频率 与 batch*PRINT_FREQ
    NUMS_VAL_DICT = {2: 1, 35: 1, 50: 1}  # 验证起始和频率
    NUMS_TEST_DICT = {2: 1, 35: 1, 50: 1}  # 测试的起始和频率
    NUM_SAVE_INTERVAL = 1  # 保存频率
    MAPS_DEF_MAX = [0.5, 0.5]  # MAP的最大值 触发保存

    # 数据集
    # BATCH_TRAIN = 3
    # BATCH_VAL = 1

    # 数据增强
    IS_MULTI_SCALE = False  # 开启多尺寸训练 只要用数据增强必须有
    MULTI_SCALE_VAL = [200, 300]  # 多尺寸训练的尺寸范围

    # 其它动态参数
    PATH_SAVE_WEIGHT = ''

    def __str__(self) -> str:
        s = '---------------- CFG参数 ------------------\n'
        for name in dir(self):
            if not name.startswith('__'):
                s += ('\t' + name + ' : ' + str(getattr(self, name)) + '\n')
        return s



import math
import torch
import numpy as np


def empty_bboxes(bboxs):
    if isinstance(bboxs, np.ndarray):
        bboxs_copy = np.empty_like(bboxs, dtype=np.float)
    elif isinstance(bboxs, torch.Tensor):
        device = bboxs.device
        bboxs_copy = torch.empty_like(bboxs, device=device, dtype=torch.float)
    else:
        raise Exception('类型错误', type(bboxs))
    return bboxs_copy


def ltrb2xywh(bboxs):
    bboxs_copy = empty_bboxes(bboxs)  # 复制矩阵
    # wh = rb - lt
    bboxs_copy[..., 2:] = bboxs[..., 2:] - bboxs[..., :2]
    # xy = lt + 0.5* cwh    or (lt+rb)/2
    bboxs_copy[..., :2] = bboxs[..., :2] + 0.5 * bboxs_copy[..., 2:]
    return bboxs_copy


def ltrb2ltwh(bboxs):
    bboxs_copy = empty_bboxes(bboxs)
    # lt = lt
    bboxs_copy[..., :2] = bboxs[..., :2]
    # wh = rb -lt
    bboxs_copy[..., 2:] = bboxs[..., 2:] - bboxs[..., :2]
    return bboxs_copy


def ltwh2ltrb(bboxs):
    bboxs_copy = empty_bboxes(bboxs)
    # lt = lt
    bboxs_copy[..., :2] = bboxs[..., :2]
    # rb= lt +wh
    bboxs_copy[..., 2:] = bboxs[..., :2] + bboxs[..., 2:]
    return bboxs_copy


def xywh2ltrb(bboxs):
    bboxs_copy = empty_bboxes(bboxs)

    # if isinstance(bboxs_copy, np.ndarray):
    #     fdiv = np.true_divide
    #     v = 2
    # elif isinstance(bboxs_copy, torch.Tensor):
    #     fdiv = torch.true_divide
    #     v = torch.tensor(2, device=bboxs.device)
    # else:
    #     raise Exception('类型错误', type(bboxs))

    # lt = xy - wh/2
    bboxs_copy[..., :2] = bboxs[..., :2] - bboxs[..., 2:] * 0.5
    # rb = clt + wh
    bboxs_copy[..., 2:] = bboxs_copy[..., :2] + bboxs[..., 2:]
    return bboxs_copy


def xywh2ltwh(bboxs):
    bboxs_copy = empty_bboxes(bboxs)

    if isinstance(bboxs_copy, np.ndarray):
        fdiv = np.true_divide
        v = 2
    elif isinstance(bboxs_copy, torch.Tensor):
        fdiv = torch.true_divide
        v = torch.tensor(2, device=bboxs.device)
    else:
        raise Exception('类型错误', type(bboxs))

    # lt = xy - wh/2
    bboxs_copy[..., :2] = bboxs[..., :2] - fdiv(bboxs[..., 2:], v)
    # wh=wh
    bboxs_copy[..., 2:] = bboxs[..., 2:]
    return bboxs_copy


def ltwh2xywh(bboxs):
    bboxs_copy = empty_bboxes(bboxs)

    # xy= lt+ wh*0.5
    bboxs_copy[..., :2] = bboxs[..., :2] + bboxs[..., 2:] * 0.5
    # wh=wh
    bboxs_copy[..., 2:] = bboxs[..., 2:]
    return bboxs_copy


def bbox_iou4np(bbox_a, bbox_b):
    '''
    求所有bboxs 与所有标定框 的交并比 ltrb
    返回一个数
    :param bbox_a: 多个预测框 (n,4)
    :param bbox_b: 多个标定框 (k,4)
    :return: <class 'tuple'>: (2002, 2) (n,k)
    '''
    if bbox_a.shape[1] != 4 or bbox_b.shape[1] != 4:
        raise IndexError
    '''
    确定交叉部份的坐标  以下设 --- n = 3 , k = 4 ---
    广播 对a里每个bbox都分别和b里的每个bbox求左上角点坐标最大值 输出 (n,k,2)
    左上 右下
    '''
    tl = np.maximum(bbox_a[:, None, :2], bbox_b[:, :2])  # (3,1,2) (4,2)--->(3,4,2)
    # 选出前两个最小的 ymin，xmin 左上角的点 后面的通过广播
    br = np.minimum(bbox_a[:, None, 2:], bbox_b[:, 2:])
    '''
    相交的面积 只有当右下的最小坐标  >>(xy全大于) 左上的最大坐标时才相交 用右下-左上得长宽
    '''
    # (2002,2,2)的每一第三维 降维运算(2002,2)  通过后面的是否相交的 降维运算 (2002,2)赛选
    area_i = np.prod(br - tl, axis=2) * (tl < br).all(axis=2)
    # (2002,2) axis=1 2->1 行相乘 长*宽 --->降维运算(2002)
    area_a = np.prod(bbox_a[:, 2:] - bbox_a[:, :2], axis=1)
    # (2,2) axis=1 2->1 行相乘 长*宽 --->降维运算(2)
    area_b = np.prod(bbox_b[:, 2:] - bbox_b[:, :2], axis=1)
    #  (2002,1) +(2) = (2002,2) 每个的二维面积
    _a = area_a[:, None] + area_b
    _area_all = (_a - area_i)  # (2002,2)-(2002,2)
    return area_i / _area_all  # (2002,2)


def bbox_iou_v3(boxes1, boxes2, mode='iou', is_aligned=False):
    '''
    overlaps
    :param boxes1: 支持 [n,4]^^[n,4] 或 [n,4]^^[m,4]
    :param boxes2:
    :param mode: 'iou' 'giou'
    :param is_aligned:
    :return:
    # 3,4
    boxes1 = torch.FloatTensor([
        [0, 0, 10, 10],
        [10, 10, 20, 20],
        # [32, 32, 38, 42],
    ])

    # 2,4
    boxes2 = torch.FloatTensor([
        [0, 0, 10, 20],
        [0, 10, 10, 19],
        # [10, 10, 20, 20],
    ])
    print(bbox_overlaps(boxes1=boxes1, boxes2=boxes2, mode='giou', is_aligned=True))
    结果 -------------------
    iou
    tensor([[0.5000, 0.0000, 0.0000],
            [0.0000, 0.0000, 1.0000],
            [0.0000, 0.0000, 0.0000]])
    giou
    tensor([[0.5000, 0.0000, -0.5000],
            [-0.2500, -0.0500, 1.0000],
            [-0.8371, -0.8766, -0.8214]])
    '''

    assert mode in ['iou', 'iof', 'giou'], f'Unsupported mode {mode}'
    # Either the boxes are empty or the length of boxes's last dimenstion is 4
    assert (boxes1.size(-1) == 4 or boxes1.size(0) == 0)
    assert (boxes2.size(-1) == 4 or boxes2.size(0) == 0)

    # Batch dim must be the same
    # Batch dim: (B1, B2, ... Bn)
    assert boxes1.shape[:-2] == boxes2.shape[:-2]
    eps = torch.finfo(torch.float16).eps
    batch = boxes1.shape[:-2]

    rows = boxes1.size(-2)
    cols = boxes2.size(-2)
    if is_aligned:
        assert rows == cols

    if rows * cols == 0:
        if is_aligned:
            return boxes1.new(batch + (rows,))
        else:
            return boxes1.new(batch + (rows, cols))

    area1 = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    area2 = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

    if is_aligned:
        lt = torch.max(boxes1[..., :2], boxes2[..., :2])  # [B, rows, 2]
        rb = torch.min(boxes1[..., 2:], boxes2[..., 2:])  # [B, rows, 2]

        wh = (rb - lt).clamp(min=0)  # [B, rows, 2]
        # torch.Size([3, 2100])^^ torch.Size([3, 2100])
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1 + area2 - overlap
        else:
            union = area1
        if mode == 'giou':
            enclosed_lt = torch.min(boxes1[..., :2], boxes2[..., :2])
            enclosed_rb = torch.max(boxes1[..., 2:], boxes2[..., 2:])
    else:
        lt = torch.max(boxes1[..., :, None, :2],
                       boxes2[..., None, :, :2])  # [B, rows, cols, 2]
        rb = torch.min(boxes1[..., :, None, 2:],
                       boxes2[..., None, :, 2:])  # [B, rows, cols, 2]

        wh = (rb - lt).clamp(min=0)  # [B, rows, cols, 2]
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1[..., None] + area2[..., None, :] - overlap
        else:
            union = area1[..., None]
        if mode == 'giou':
            enclosed_lt = torch.min(boxes1[..., :, None, :2],
                                    boxes2[..., None, :, :2])
            enclosed_rb = torch.max(boxes1[..., :, None, 2:],
                                    boxes2[..., None, :, 2:])

    eps = union.new_tensor([eps])
    union = torch.max(union, eps)
    ious = overlap / union
    if mode in ['iou', 'iof']:
        return ious
    # calculate gious
    enclose_wh = (enclosed_rb - enclosed_lt).clamp(min=0)
    enclose_area = enclose_wh[..., 0] * enclose_wh[..., 1]
    enclose_area = torch.max(enclose_area, eps)
    gious = ious - (enclose_area - union) / enclose_area
    return gious


def bbox_iou4aligned(box1_ltrb, box2_ltrb, is_giou=False, is_diou=False, is_ciou=False):
    '''
    这个是一一对应  box1  box2 支持2D和3D
    :param box1_ltrb:
    :param box2_ltrb:
    iou: 目标框不相交时全为0 不能反映如何相交的  --- 重叠面积
    giou: 当目标框完全包裹预测框的时候 退化为iou   --- 中心点距离
    diou: 提高收敛速度及全包裹优化
    ciou: --- 长宽比  CIOU有点不一致
    :return: n
    '''
    eps = torch.finfo(torch.float16).eps
    # torch.Size([2, 845, 2])
    max_lt = torch.max(box1_ltrb[..., :2], box2_ltrb[..., :2])  # left-top [N,M,2] 多维组合用法
    min_rb = torch.min(box1_ltrb[..., 2:], box2_ltrb[..., 2:])  # right-bottom [N,M,2]
    # 确保面积不能为负  否则以0计算
    inter_wh = (min_rb - max_lt).clamp(min=0)  # [N,M,2]
    # inter_wh = (min_rb - max_lt).clamp(min=0)  # [N,M,2]
    # inter_area = inter_wh[:, 0] * inter_wh[:, 1]  # [N,M] 降维
    inter_area = torch.prod(inter_wh, dim=-1)  # 这个一定>=0

    # 并的面积 单项面积为负 则置0
    # area1 = (box1_ltrb[..., 2] - box1_ltrb[..., 0]) * (box1_ltrb[..., 3] - box1_ltrb[..., 1]).clamp(0)
    # area2 = (box2_ltrb[..., 2] - box2_ltrb[..., 0]) * (box2_ltrb[..., 3] - box2_ltrb[..., 1]).clamp(0)
    area1 = torch.prod(box1_ltrb[..., 2:] - box1_ltrb[..., :2], -1).clamp(0)
    area2 = torch.prod(box2_ltrb[..., 2:] - box2_ltrb[..., :2], -1).clamp(0)
    ''' 确保这个不为负 '''
    union_area = (area1 + area2 - inter_area).clamp(eps)  # A+B-交=并

    iou = inter_area / union_area  # 交一定小于并
    # flog.debug('iou %s', iou)

    if not (is_giou or is_diou or is_ciou):
        return iou
    else:
        # 最大矩形面积
        min_lt = torch.min(box1_ltrb[..., :2], box2_ltrb[..., :2])
        max_rb = torch.max(box1_ltrb[..., 2:], box2_ltrb[..., 2:])
        max_wh = max_rb - min_lt
        if is_giou:
            ''' 确保这个不为负 '''
            max_area = (max_wh[..., 0] * max_wh[..., 1]).clamp(eps)
            giou = iou - (max_area - union_area) / max_area
            return giou

        ''' 确保这个不为负 '''
        c2 = (max_wh[..., 0] ** 2 + max_wh[..., 1] ** 2).clamp(eps)  # 最大矩形的矩离的平方
        box1_xywh = ltrb2xywh(box1_ltrb)
        box2_xywh = ltrb2xywh(box2_ltrb)
        xw2_xh2 = torch.pow(box1_xywh[..., :2] - box2_xywh[..., :2], 2)  # 中心点距离的平方
        d2 = xw2_xh2[..., 0] + xw2_xh2[..., 1]
        dxy = d2 / c2  # 中心比例距离
        if is_diou:
            diou = iou - dxy
            return diou

        if is_ciou:
            # [3, 1] / [3, 1]  => [3, 1]
            box1_atan_wh = torch.atan(box1_xywh[..., 2:3] / box1_xywh[..., 3:])
            box2_atan_wh = torch.atan(box2_xywh[..., 2:3] / box2_xywh[..., 3:])
            # torch.Size([3, 1])
            v = torch.pow(box1_atan_wh - box2_atan_wh, 2) * (4 / math.pi ** 2)
            v = torch.squeeze(v, -1)  # m,n,1 -> m,n 去掉最后一维
            # v.squeeze_(-1)  # m,n,1 -> m,n 去掉最后一维
            with torch.no_grad():
                alpha = v / (1 - iou + v)
            ciou = iou - (dxy + v * alpha)
            return ciou


def bbox_iou4unaligned(boxes1, boxes2, is_giou=False, is_diou=False, is_ciou=False):
    '''

    :param boxes1:torch.Size([m, 4]) ltrb
    :param boxes2:torch.Size([n, 4]) ltrb
    :param is_giou: 重合度++
    :param is_diou: +中心点
    :param is_ciou: +宽高
    :return:(m,n)
    '''
    # 交集面积
    max_lt = torch.max(boxes1[:, None, :2], boxes2[:, :2])  # left-top [N,M,2] 多维组合用法
    min_rb = torch.min(boxes1[:, None, 2:], boxes2[:, 2:])  # right-bottom [N,M,2]
    inter_wh = (min_rb - max_lt).clamp(min=0)  # [N,M,2]
    inter_area = inter_wh[:, :, 0] * inter_wh[:, :, 1]  # [N,M] 降维

    # 并的面积
    area1 = bbox_area(boxes1)  # 降维 n
    area2 = bbox_area(boxes2)  # 降维 m
    union_area = area1[:, None] + area2 - inter_area + torch.finfo(torch.float16).eps  # 升维n m

    iou = inter_area / union_area

    if not (is_giou or is_diou or is_ciou):
        return iou
    else:
        # 最大矩形面积
        min_lt = torch.min(boxes1[:, None, :2], boxes2[:, :2])
        max_rb = torch.max(boxes1[:, None, 2:], boxes2[:, 2:])
        max_wh = max_rb - min_lt
        if is_giou:
            max_area = max_wh[:, :, 0] * max_wh[:, :, 1] + torch.finfo(torch.float16).eps  # 降维运算
            giou = iou - (max_area - union_area) / max_area
            return giou

        c2 = max_wh[:, :, 0] ** 2 + max_wh[:, :, 1] ** 2 + torch.finfo(torch.float16).eps  # 最大矩形的矩离的平方
        box1_xywh = ltrb2xywh(boxes1)
        box2_xywh = ltrb2xywh(boxes2)
        xw2_xh2 = torch.pow(box1_xywh[:, None, :2] - box2_xywh[:, :2], 2)  # 中心点距离的平方
        d2 = xw2_xh2[:, :, 0] + xw2_xh2[:, :, 1]
        dxy = d2 / c2  # 中心比例距离
        if is_diou:
            diou = iou - dxy
            return diou

        if is_ciou:
            box1_atan_wh = torch.atan(box1_xywh[:, 2:3] / box1_xywh[:, 3:])  # w/h
            box2_atan_wh = torch.atan(box2_xywh[:, 2:3] / box2_xywh[:, 3:])
            # torch.squeeze(ts)
            v = torch.pow(box1_atan_wh[:, None, :] - box2_atan_wh, 2) * (4 / math.pi ** 2)
            # v = torch.squeeze(v, -1)  # m,n,1 -> m,n 去掉最后一维
            v.squeeze_(-1)  # m,n,1 -> m,n 去掉最后一维
            with torch.no_grad():
                alpha = v / (1 - iou + v)
            ciou = iou - (dxy + v * alpha)
            return ciou


def bbox_iou4fcos(poff_ltrb_exp, goff_ltrb):
    w_gt = goff_ltrb[:, :, 0] + goff_ltrb[:, :, 2]
    h_gt = goff_ltrb[:, :, 1] + goff_ltrb[:, :, 3]
    w_pred = poff_ltrb_exp[:, :, 0] + poff_ltrb_exp[:, :, 2]
    h_pred = poff_ltrb_exp[:, :, 1] + poff_ltrb_exp[:, :, 3]
    S_gt = w_gt * h_gt
    S_pred = w_pred * h_pred
    I_h = torch.min(goff_ltrb[:, :, 1], poff_ltrb_exp[:, :, 1]) + torch.min(goff_ltrb[:, :, 3], poff_ltrb_exp[:, :, 3])
    I_w = torch.min(goff_ltrb[:, :, 0], poff_ltrb_exp[:, :, 0]) + torch.min(goff_ltrb[:, :, 2], poff_ltrb_exp[:, :, 2])
    S_I = I_h * I_w
    U = S_gt + S_pred - S_I + 1e-20
    IoU = S_I / U

    return IoU


def bbox_area(boxes):
    ''' boxes 一定有个面积 '''
    return torch.abs(boxes[..., 2] - boxes[..., 0]) * (boxes[..., 3] - boxes[..., 1])


if __name__ == '__main__':
    # 3,4
    boxes1 = torch.FloatTensor([
        [0, 0, 10, 10],
        [10, 10, 20, 20],
        # [32, 32, 38, 42],
    ])

    # 2,4
    boxes2 = torch.FloatTensor([
        [0, 0, 10, 20],
        [0, 10, 10, 19],
        [10, 10, 20, 20],
    ])
    print(bbox_iou_v3(boxes1=boxes1, boxes2=boxes2, mode='giou', is_aligned=False))
    print(bbox_iou4unaligned(boxes1=boxes1, boxes2=boxes2, is_giou=True))



import warnings
from abc import ABCMeta

import torch

from tools.GLOBAL_LOG import flog
from tools.boxes.f_boxes import bbox_iou_v3, ltrb2xywh, xywh2ltrb
from tools.datas.f_data_pretreatment4np import f_recover_normalization4ts
from tools.f_calc_adv import f_cre_grid_cells


def match4atss(gltrb_i_b, anc_ltrb_i, nums_dim_t_list, glabels_b=None,
               img_ts=None, is_visual=False):
    '''
    这里默认都是 input 尺寸 且是batch
    核心作用:
        1. 使正例框数量都保持一致 保障小目标也能匹配到多个anc
        2. 用最大一个iou的均值和标准差,计算阀值,用IOU阀值初选正样本
        3. 确保anc中心点在gt框中

    :param gltrb_i_b:
    :param anc_ltrb_i:
    :param nums_dim_t_list:
        [1600, 400, 100] 这个是每个特图对应的维度数 用于索引 如果只有一层可以优化掉
    :param glabels_b: 暂时没用 可视化
    :param img_ts: 可视化
    :param is_visual:  可视化
    :return:
        mask_pos : [2100] 正例mask
        anc_max_iou: [2100] anc 对应的最大IOU值
        g_pos_index: [2100] anc 对应GT的索引
    '''
    device = gltrb_i_b.device

    # 计算 iou
    anc_xywh_i = ltrb2xywh(anc_ltrb_i)
    # num_anc = anc_xywh_i.shape[0]
    # (anc 个,boxes 个) torch.Size([3, 10647])
    ious_ag = bbox_iou_v3(anc_ltrb_i, gltrb_i_b)
    num_gt = gltrb_i_b.shape[0]  # 正样本个数

    # 全部ANC的距离
    gxywh_i_b = ltrb2xywh(gltrb_i_b)
    # 中间点绝对距离 多维广播 (anc 个,boxes 个)  torch.Size([32526, 7])
    distances = (anc_xywh_i[:, None, :2] - gxywh_i_b[None, :, :2]).pow(2).sum(-1).sqrt()

    # 每层 anc 数是一致的
    num_atss_topk = 9  # 这个 topk = 要 * 该层的anc数

    idxs_candidate = []  # 这个用来保存最一层所匹配的最小距离anc的索引  每层9个
    index_start = 0  # 这是每层的anc偏移值
    for i, num_dim_feature in enumerate(nums_dim_t_list):  # [24336, 6084, 1521, 441, 144]
        '''每一层的每一个GT选 topk * anc数'''
        index_end = index_start + num_dim_feature
        # 取出某层的所有anc距离  中间点绝对距离 (anc 个,boxes 个)  torch.Size([32526, 7]) -> [nn, 7]
        distances_per_level = distances[index_start:index_end, :]
        # 确认该层的TOPK 不能超过该层总 anc 数 这里是一个数
        topk = min(num_atss_topk, num_dim_feature)
        # 选 topk个最小的 每个gt对应对的anc的index torch.Size([24336, box_n])---(anc,gt) -> torch.Size([topk, 1])
        _, topk_idxs_per_level = distances_per_level.topk(topk, dim=0, largest=False)  # 只能在某一维top
        idxs_candidate.append(topk_idxs_per_level + index_start)
        index_start = index_end

    # 用于计算iou均值和方差 候选人，候补者；应试者 torch.Size([405, 1])
    idxs_candidate = torch.cat(idxs_candidate, dim=0)
    '''--- 选出每层每个anc对应的距离中心最近topk iou值 ---'''
    # ***************这个是ids选择 这个是多维筛选 ious---[anc,ngt]    [405, ngt] [0,1...ngt]-> [405,ngt]
    ious_candidate = ious_ag[idxs_candidate, torch.arange(num_gt)]  # 这里是index 蒙板取数的方法
    mask_distances = torch.zeros_like(distances, device=device, dtype=torch.bool)
    # [2000,ngt]
    mask_distances[idxs_candidate, torch.arange(idxs_candidate.shape[1])] = True

    if is_visual:
        # ***********  debug 可视  匹配29个正例可视化  *****************
        from tools.picture.f_show import f_show_od_ts4plt_v3
        # 多个gt 对应一个anc 进行转换显示
        mask_pos = torch.zeros(mask_distances.shape[0], device=device, dtype=torch.bool)
        mask_distances_t = mask_distances.t()  # [32526 , 3] -> [3, 32526]
        for m_pos_iou in mask_distances_t:  # 拉平 每个 32526
            # 遍历GT 直接进行一个或操作 更新 mask
            mask_pos = torch.logical_or(m_pos_iou, mask_pos)
        f_show_od_ts4plt_v3(img_ts, g_ltrb=gltrb_i_b,
                            p_ltrb=anc_ltrb_i[mask_pos],
                            is_normal=True,
                            )

    '''--- 用最大一个iou的均值和标准差,计算阀值 ---'''
    # 统计每一个 GT的均值 std [ntopk,ngt] -> [ngt] 个
    _iou_mean_per_gt = ious_candidate.mean(dim=0)  # 除维
    _iou_std_per_gt = ious_candidate.std(dim=0)
    _iou_thresh_per_gt = _iou_mean_per_gt + _iou_std_per_gt
    '''--- 用IOU阀值初选正样本 ---'''
    # torch.Size([32526, 1]) ^^ ([ngt] -> [1,ngt]) -> [32526,ngt]
    mask_pos4iou = ious_ag >= _iou_thresh_per_gt[None, :]  # 核心是这个选

    '''--- 中心点需落在GT中间 需要选出 anc的中心点-gt的lt为正, gr的rb-anc的中心点为正  ---'''
    # torch.Size([32526, 1, 2])
    dlt = anc_xywh_i[:, None, :2] - gltrb_i_b[None, :, :2]
    drb = gltrb_i_b[None, :, 2:] - anc_xywh_i[:, None, :2]
    # [32526, 1, 2] -> [32526, 1, 4] -> [32526, 1]
    mask_pos4in_gt = torch.all(torch.cat([dlt, drb], dim=-1) > 0.01, dim=-1)
    mask_pos_iou = torch.logical_and(torch.logical_and(mask_distances, mask_pos4iou), mask_pos4in_gt)

    '''--- 生成最终正例mask [32526, ngt] -> [32526] ---'''
    mask_pos = torch.zeros(mask_pos_iou.shape[0], device=device, dtype=torch.bool)
    mask_pos_iou = mask_pos_iou.t()  # [32526 , 3] -> [3, 32526]
    for m_pos_iou in mask_pos_iou:  # 拉平 每个 32526
        mask_pos = torch.logical_or(m_pos_iou, mask_pos)

    '''--- 确定anc匹配 一个锚框被多个真实框所选择，则其归于iou较高的真实框  ---'''
    # (anc 个,boxes 个) torch.Size([3, 10647])
    anc_max_iou, g_index = ious_ag.max(dim=1)  # 存的是 bboxs的index

    if is_visual:
        # ***********  debug 可视  多重过滤后个正例可视化  *****************
        from tools.picture.f_show import f_show_od_ts4plt_v3
        f_show_od_ts4plt_v3(img_ts, g_ltrb=gltrb_i_b,
                            p_ltrb=anc_ltrb_i[mask_pos],
                            is_normal=True,
                            )

    return mask_pos, anc_max_iou, g_index


def decode4nanodet(grids_xy_input, p_tltrb, max_size_hw=None):
    '''

    :param grids_xy_input: torch.Size([2100, 2])
    :param p_tltrb: torch.Size([3, 2100, 4])
    :param max_size_hw: 预测时使用
    :return:
    '''
    # torch.Size([3, 2100])
    x1 = grids_xy_input[:, 0] - p_tltrb[..., 0]
    y1 = grids_xy_input[:, 1] - p_tltrb[..., 1]
    x2 = grids_xy_input[:, 0] + p_tltrb[..., 2]
    y2 = grids_xy_input[:, 1] + p_tltrb[..., 3]
    if max_size_hw is not None:
        x1 = x1.clamp(min=0, max=max_size_hw[1])
        y1 = y1.clamp(min=0, max=max_size_hw[0])
        x2 = x2.clamp(min=0, max=max_size_hw[1])
        y2 = y2.clamp(min=0, max=max_size_hw[0])
    # torch.Size([3, 2100]) x4 torch.Size([3, 2100,4])
    return torch.stack([x1, y1, x2, y2], -1)


def encode4nanodet(grids_xy_input, g_ltrb, max_val, eps=0.1, mask_debug=None):
    '''
    编码针对特图
    :param grids_xy_input:  torch.Size([2100, 2])
    :param g_ltrb:  torch.Size([3, 2100, 4])
    :param max_val: 限制匹配的正例 最大距离 在0~7 之内
    :param mask_debug:  用于查看 GT 与 匹配的点 的ltrb距离是否会超过7
    :param eps:
    :return:
    '''
    from tools.GLOBAL_LOG import flog
    left = grids_xy_input[:, 0] - g_ltrb[..., 0]
    top = grids_xy_input[:, 1] - g_ltrb[..., 1]
    right = g_ltrb[..., 2] - grids_xy_input[:, 0]
    bottom = g_ltrb[..., 3] - grids_xy_input[:, 1]
    g_tltrb = torch.stack([left, top, right, bottom], -1)
    flog.debug('正例 g_ltrb min=%f  max=%f' % (g_tltrb[mask_debug].min(), g_tltrb[mask_debug].max()))
    if max_val is not None:
        g_tltrb = g_tltrb.clamp(min=0, max=max_val - eps)
    return g_tltrb


def match_yolo1_od(g_ltrb_input_b, g_labels_b, size_wh_t_ts, device, cfg, img_ts_input, toone_wh_ts_input=None):
    '''
    匹配 gyolo 如果需要计算IOU 需在这里生成
    :param g_ltrb_input_b: ltrb
    :param g_labels_b:
    :param grid: 13
    :param device:
    :return:
    '''
    num_gt = g_ltrb_input_b.shape[0]
    g_txywh_t, weights, indexs_colrow_t = encode_yolo1_od(g_ltrb_input_b, size_wh_t_ts, cfg)

    g_cls_b_ = torch.zeros((size_wh_t_ts[1], size_wh_t_ts[0], cfg.NUM_CLASSES), device=device)
    g_weight_b_ = torch.zeros((size_wh_t_ts[1], size_wh_t_ts[0], 1), device=device)
    g_txywh_t_b_ = torch.zeros((size_wh_t_ts[1], size_wh_t_ts[0], 4), device=device)

    labels_index = (g_labels_b - 1).long()
    indexs_colrow_t = indexs_colrow_t.long()  # index需要long类型

    for i in range(num_gt):
        g_cls_b_[indexs_colrow_t[i, 1], indexs_colrow_t[i, 0], labels_index[i]] = 1  # 构建 onehot
        g_weight_b_[indexs_colrow_t[i, 1], indexs_colrow_t[i, 0]] = weights[i]
        g_txywh_t_b_[indexs_colrow_t[i, 1], indexs_colrow_t[i, 0]] = g_txywh_t[i]

    g_cls_b_ = g_cls_b_.reshape(-1, cfg.NUM_CLASSES)
    g_weight_b_ = g_weight_b_.reshape(-1, 1)
    g_txywh_t_b_ = g_txywh_t_b_.reshape(-1, 4)

    '''可视化验证'''
    if cfg.IS_VISUAL:
        mask_pos_1d = (g_weight_b_ > 0).any(-1)  # [169]

        # [2] -> [1,2] -> [ngt,2]
        sizes_wh_t_ts = size_wh_t_ts.unsqueeze(0).repeat(num_gt, 1)

        flog.debug('size_wh_t_ts = %s', size_wh_t_ts)
        flog.debug('toone_wh_ts_input = %s', toone_wh_ts_input)
        flog.debug('g_txywh_t = %s', g_txywh_t)
        flog.debug('indexs_colrow_t = %s', indexs_colrow_t)
        flog.debug('对应index = %s', indexs_colrow_t[0, 1] * size_wh_t_ts[0] + indexs_colrow_t[0, 0])
        flog.debug('对应index = %s', torch.where(g_weight_b_ > 0))

        flog.debug('g_ltrb_input_b = %s', g_ltrb_input_b)

        grid_wh = [size_wh_t_ts[0].item(), size_wh_t_ts[1].item()]
        # 预测时需要sigmoid  这里的 g_txywh_t_b_ 已 sigmoid
        p_ltrb_t = decode_yolo1_od(p_txywh_t_sigmoidxy=g_txywh_t_b_, grid_wh=grid_wh,
                                   toones_wh_ts_t=sizes_wh_t_ts, is_to_one=False)
        p_ltrb_t_pos = p_ltrb_t[mask_pos_1d]

        from f_tools.pic.enhance.f_data_pretreatment4np import f_recover_normalization4ts

        img_ts = f_recover_normalization4ts(img_ts_input)
        from torchvision.transforms import functional as transformsF
        img_pil = transformsF.to_pil_image(img_ts).convert('RGB')
        import numpy as np
        img_np = np.array(img_pil)

        # 特图 -> input
        # img_wh_ts_x2 = torch.tensor(img_np.shape[:2][::-1], device=device).repeat(2)
        p_ltrb_input_pos = p_ltrb_t_pos * cfg.STRIDE

        flog.debug('p_ltrb_input_pos = %s', p_ltrb_input_pos)
        flog.debug(' ----------------------------------------- ')
        f_show_od_np4plt_v2(img_np,
                            gboxes_ltrb=g_ltrb_input_b.cpu(),
                            pboxes_ltrb=p_ltrb_input_pos.cpu(),
                            is_recover_size=False,
                            grid_wh_np=size_wh_t_ts.cpu().numpy())

    return g_cls_b_, g_weight_b_, g_txywh_t_b_


def encode_yolo1_od(g_ltrb_input_b, size_wh_t_ts, cfg):
    # ltrb -> xywh 原图归一化   编码xy与yolo2一样的
    g_xywh_input = ltrb2xywh(g_ltrb_input_b)
    g_xywh_t = g_xywh_input / cfg.STRIDE
    cxys_t = g_xywh_t[:, :2]
    whs_t = g_xywh_t[:, 2:]
    whs_one = whs_t / size_wh_t_ts

    # 转换到特图的格子中
    indexs_colrow_t = cxys_t.floor()
    g_txy_t = cxys_t - indexs_colrow_t
    g_twh_t = whs_t.log()
    g_txywh_t = torch.cat([g_txy_t, g_twh_t], dim=-1)

    # 值在 1~2 之间 放大小的
    weights = 2.0 - torch.prod(whs_one, dim=-1)

    return g_txywh_t, weights, indexs_colrow_t


def decode_yolo1_od(p_txywh_t_sigmoidxy, grid_wh, toones_wh_ts_t, is_to_one):
    '''
    解码出来是特图
    :param p_txywh_t_sigmoidxy:  必须 c 是最后一维
    :return: 输出原图归一化
    '''
    device = p_txywh_t_sigmoidxy.device
    # 单一格子偏移 + 特图格子偏移
    p_xy_t = p_txywh_t_sigmoidxy[..., :2] + f_cre_grid_cells((grid_wh[1], grid_wh[0]), is_swap=True, num_repeat=1).to(
        device)
    p_wh_t = p_txywh_t_sigmoidxy[..., 2:].exp()
    p_xywh_t = torch.cat([p_xy_t, p_wh_t], -1)
    p_ltrb_t = xywh2ltrb(p_xywh_t)
    if is_to_one:
        # torch.Size([2, 169, 4])  ^^ batch,1,4
        p_ltrb_one = torch.true_divide(p_ltrb_t, toones_wh_ts_t.repeat(1, 2).unsqueeze(1))  # 特图-》 one
        return p_ltrb_one
    return p_ltrb_t



from abc import abstractmethod

import torch
from torch import nn


def nms(boxes, scores, iou_threshold):
    ''' IOU大于0.5的抑制掉
         boxes (Tensor[N, 4])) – bounding boxes坐标. 格式：(ltrb
         scores (Tensor[N]) – bounding boxes得分
         iou_threshold (float) – IoU过滤阈值
     返回:NMS过滤后的bouding boxes索引（降序排列）
     '''
    return torch.ops.torchvision.nms(boxes, scores, iou_threshold)


def batched_nms(boxes_ltrb, scores, idxs, threshold_iou):
    '''
    多 labels nms
    :param boxes_ltrb: 拉平所有类别的box重复的 n*20类,4
    :param scores: torch.Size([16766])
    :param idxs:  真实类别index 通过手动创建匹配的 用于表示当前 nms的类别 用于统一偏移 技巧
    :param threshold_iou:float 0.5
    :return:
    '''
    if boxes_ltrb.numel() == 0:  # 维度全乘
        return torch.empty((0,), dtype=torch.int64, device=boxes_ltrb.device)

    # torchvision.ops.boxes.batched_nms(boxes, scores, lvl, nms_thresh)
    # 根据最大的一个值确定每一类的偏移
    max_coordinate = boxes_ltrb.max()  # 选出每个框的 坐标最大的一个值
    # idxs 的设备和 boxes 一致 , 真实类别index * (1+最大值) 则确保同类框向 左右平移 实现隔离
    offsets = idxs.to(boxes_ltrb) * (max_coordinate + 1)
    # boxes 加上对应层的偏移量后，保证不同类别之间boxes不会有重合的现象
    boxes_for_nms = boxes_ltrb + offsets[:, None]
    keep = nms(boxes_for_nms, scores, threshold_iou)
    return keep


def batch_nms(ids_batch1, p_boxes_ltrb1, p_labels1, p_scores1, threshold_nms):
    '''
    以每一个batch 为index 对 每一类进行偏移
    :param ids_batch1: [nn]
    :param p_boxes_ltrb1: [nn,4] float32
    :param p_labels1: [nn]
    :param p_scores1: [nn]
    :param threshold_nms:
    :return:
    '''
    device = p_boxes_ltrb1.device
    p_labels_unique = p_labels1.unique()  # nn -> n
    ids_batch2 = torch.tensor([], device=device)
    p_scores2 = torch.tensor([], device=device)
    p_labels2 = torch.tensor([], device=device)
    p_boxes_ltrb2 = torch.empty((0, 4), dtype=torch.float, device=device)

    for lu in p_labels_unique:  # 必须每类处理
        # 过滤类别
        _mask = p_labels1 == lu
        _ids_batch = ids_batch1[_mask]
        _p_scores = p_scores1[_mask]
        _p_labels = p_labels1[_mask]
        _p_boxes_ltrb = p_boxes_ltrb1[_mask]
        keep = batched_nms(_p_boxes_ltrb, _p_scores, _ids_batch, threshold_nms)
        # 极大抑制
        ids_batch2 = torch.cat([ids_batch2, _ids_batch[keep]])
        p_scores2 = torch.cat([p_scores2, _p_scores[keep]])
        p_labels2 = torch.cat([p_labels2, _p_labels[keep]])
        p_boxes_ltrb2 = torch.cat([p_boxes_ltrb2, _p_boxes_ltrb[keep]])
        # print('12')
    return ids_batch2, p_boxes_ltrb2, p_labels2, p_scores2


def batch_nms4kp(ids_batch1, p_boxes_ltrb1, p_labels1, p_scores1, p_keypoints1, threshold_nms):
    '''
    分类 nms
    :param ids_batch1: [nn]
    :param p_boxes_ltrb1: [nn,4] float32
    :param p_labels1: [nn]
    :param p_scores1: [nn]
    :param device:
    :param threshold_nms:
    :return:
    '''
    p_labels_unique = p_labels1.unique()  # nn -> n
    ids_batch2 = torch.tensor([], device=device)
    p_scores2 = torch.tensor([], device=device)
    p_labels2 = torch.tensor([], device=device)
    p_boxes_ltrb2 = torch.empty((0, 4), dtype=torch.float, device=device)
    # 这个不需要数量
    p_keypoints2 = torch.empty((0, cfg.NUM_KEYPOINTS * 2), dtype=torch.float, device=device)

    for lu in p_labels_unique:  # 必须每类处理
        # 过滤类别
        _mask = p_labels1 == lu
        _ids_batch = ids_batch1[_mask]
        _p_scores = p_scores1[_mask]
        _p_labels = p_labels1[_mask]
        _p_boxes_ltrb = p_boxes_ltrb1[_mask]
        keep = batched_nms(_p_boxes_ltrb, _p_scores, _ids_batch, threshold_nms)
        # 极大抑制
        ids_batch2 = torch.cat([ids_batch2, _ids_batch[keep]])
        p_scores2 = torch.cat([p_scores2, _p_scores[keep]])
        p_labels2 = torch.cat([p_labels2, _p_labels[keep]])
        p_boxes_ltrb2 = torch.cat([p_boxes_ltrb2, _p_boxes_ltrb[keep]])
        if p_keypoints1 is not None:
            _p_keypoints = p_keypoints1[_mask]
            p_keypoints2 = torch.cat([p_keypoints2, _p_keypoints[keep]])
        else:
            p_keypoints2 = None
        # print('12')
    return ids_batch2, p_boxes_ltrb2, p_keypoints2, p_labels2, p_scores2,


class Predicting_Base(nn.Module):
    '''
    由 f_fit_eval_base 运行至net模型主Center 再由模型主Center调用
    返回到 f_fit_eval_base
    '''

    def __init__(self, mode_vis, thr_pred_conf=0.05,
                 thr_pred_nms=0.5, select_topk=500) -> None:
        '''

        :param mode_vis:  这里是nms的 模式 bbox keypoints
        :param thr_pred_conf:
        :param thr_pred_nms:
        :param select_topk:
        '''
        super(Predicting_Base, self).__init__()
        self.mode_vis = mode_vis
        self.thr_pred_conf = thr_pred_conf  # 阀值
        self.thr_pred_nms = thr_pred_nms  # 阀值
        self.select_topk = select_topk  # 初选个数

    @abstractmethod
    def get_pscores(self, model_outs):
        '''

        :param model_outs: p_init 处理后
        :return:
            pscores : 这个用于判断 生成 mask_pos
            plabels : 这个用于二阶段
            pconf : 用于显示conf的统计值
        '''

    @abstractmethod
    def get_stage_res(self, model_outs, mask_pos, pscores, plabels, imgs_ts_4d, targets, toones_wh_ts_input):
        '''

        :param model_outs:
        :param mask_pos:
        :param pscores:
        :param plabels:
        :param imgs_ts_4d:
        :param targets:
        :param toones_wh_ts_input: 实际归一化尺寸
        :return:
            ids_batch1 : 返回第二阶段
            pboxes_ltrb1 :
            plabels1 :
            pscores1 :
        '''

    def forward(self, model_outs, imgs_ts_4d=None, targets=None, toones_wh_ts_input=None):
        '''
        由 FModelBase 已处理  toones_wh_ts_input
        :param model_outs: 模型的输出
        :return:
        '''

        ''' 进行数据的处理 传递到 get_pscores get_stage_res 中'''
        model_outs = self.p_init(model_outs)

        ''' 返回分数 和 plabels    pconf可返回任意值 用于没有目标时分析值域 '''
        pscores, plabels, pconf = self.get_pscores(model_outs)

        mask_pos = pscores > self.thr_pred_conf
        if not torch.any(mask_pos):  # 如果没有一个对象
            print('该批次没有找到目标 max:{0:.2f} min:{0:.2f} mean:{0:.2f}'.format(pconf.max().item(),
                                                                          pconf.min().item(),
                                                                          pconf.mean().item(),
                                                                          ))
            return [None] * 5

        if pscores.shape[-1] > self.select_topk:  # 并行取top100 与mask_pos 进行and操作
            # 最大1000个
            ids_topk = pscores.topk(self.select_topk, dim=-1)[1]  # torch.Size([32, 1000])
            mask_topk = torch.zeros_like(mask_pos)
            mask_topk[torch.arange(ids_topk.shape[0])[:, None], ids_topk] = True
            mask_pos = torch.logical_and(mask_pos, mask_topk)

        # ids_batch1, pboxes_ltrb1, plabels1, pscores1,pkp_keypoint
        reses = self.get_stage_res(model_outs, mask_pos, pscores, plabels, imgs_ts_4d, targets, toones_wh_ts_input)

        if self.mode_vis == 'bbox':  # 单人脸
            assert len(reses) == 4, 'res = self.get_stage_res(outs, mask_pos, pscores, plabels) 数据错误'
            ids_batch2, p_boxes_ltrb2, p_labels2, p_scores2 = batch_nms(reses[0],
                                                                        reses[1],
                                                                        reses[2],
                                                                        reses[3],
                                                                        self.thr_pred_nms)
            return ids_batch2, p_boxes_ltrb2, None, p_labels2, p_scores2,
        elif self.mode_vis == 'keypoints':
            assert len(reses) == 5, 'res = self.get_stage_res(outs, mask_pos, pscores, plabels) 数据错误'
            ids_batch2, p_boxes_ltrb2, p_keypoints2, p_labels2, p_scores2 = batch_nms4kp(
                ids_batch1=reses[0],
                p_boxes_ltrb1=reses[1],
                p_labels1=reses[2],
                p_scores1=reses[3],
                p_keypoints1=reses[4],
                threshold_nms=self.thr_pred_nms)
            return ids_batch2, p_boxes_ltrb2, p_keypoints2, p_labels2, p_scores2
        else:
            raise Exception('self.mode_vis = %s 错误' % self.mode_vis)



import torch
import torch.nn.functional as F
import numpy as np

from tools.datas.dataset_coco import CustomCocoDataset


class CLS4collate_fn:
    # fdatas_l1
    def __init__(self, is_multi_scale):
        self.is_multi_scale = is_multi_scale

    def __call__(self, batch_datas):
        '''
        在这里重写collate_fn函数
        batch_datas: tuple[[tensor_img,dict_targets],...,[tensor_img,dict_targets]]
        '''
        # 训练才进这里
        if self.is_multi_scale:
            batch = len(batch_datas)
            imgs_list = []
            targets_list = []
            # 打开 tuple 数据
            for i, (img_ts, target) in enumerate(batch_datas):
                # flog.warning('fun4dataloader测试  %s %s %s ', target, len(target['boxes']), len(target['labels']))
                imgs_list.append(img_ts)
                targets_list.append(target)

            pad_imgs_list = []

            # 这里的最大一定能被32整除
            h_list = [int(s.shape[1]) for s in imgs_list]
            w_list = [int(s.shape[2]) for s in imgs_list]
            max_h = np.array(h_list).max()
            max_w = np.array(w_list).max()
            # self.cfg.tcfg_batch_size = [max_w, max_h] # 这样用多进程要报错
            for i in range(batch):
                img_ts = imgs_list[i]
                # 右下角添加 target 无需处理
                img_ts_pad = F.pad(img_ts, (0, int(max_w - img_ts.shape[2]), 0, int(max_h - img_ts.shape[1])), value=0.)
                pad_imgs_list.append(img_ts_pad)

                # debug 代码
                # fshow_pic_ts4plt(pad_img)  # 可视化 前面不要归一化
                # fshow_kp_ts4plt(pad_img,
                #                 targets_list[i]['boxes'],
                #                 targets_list[i]['keypoints'],
                #                 mask_kps=targets_list[i]['kps_mask'],
                #                 is_recover_size=False
                #                 )  # 可视化
                # f_show_od_ts4plt(img_ts_pad, targets_list[i]['boxes'], is_recover_size=False)
                # print('多尺度%s ' % str(pad_img.shape))

            imgs_ts_4d = torch.stack(pad_imgs_list)
        else:
            imgs_ts_3d = batch_datas[0][0]
            # 包装整个图片数据集 (batch,3,416,416) 转换到显卡
            imgs_ts_4d = torch.empty((len(batch_datas), *imgs_ts_3d.shape)).to(imgs_ts_3d)
            targets_list = []
            for i, (img, target) in enumerate(batch_datas):
                # flog.warning('fun4dataloader测试  %s %s %s ', target, len(target['boxes']), len(target['labels']))
                imgs_ts_4d[i] = img
                targets_list.append(target)
        return imgs_ts_4d, targets_list


def cre_dataloader(path_img_train=None, file_json_train=None, transform_train=None, mode_train=None, batch_train=None,
                   path_img_val=None, file_json_val=None, transform_val=None, mode_val=None, batch_val=None,
                   path_img_test=None, file_json_test=None, transform_test=None, mode_test=None, batch_test=None,
                   is_multi_scale=False, num_workers=None,
                   ):
    dataloader_train, dataloader_val, dataloader_test = None, None, None
    if file_json_train is not None:
        dataloader_train = _cre_data_load(batch_train, file_json_train, is_multi_scale,
                                          mode_train, num_workers, path_img_train,
                                          transform_train)
    if file_json_val is not None:
        dataloader_val = _cre_data_load(batch_val, file_json_val, is_multi_scale,
                                        mode_val, num_workers, path_img_val,
                                        transform_val)
    if file_json_test is not None:
        dataloader_test = _cre_data_load(batch_test, file_json_test, is_multi_scale,
                                         mode_test, num_workers, path_img_test,
                                         transform_test)
    return dataloader_train, dataloader_val, dataloader_test


def _cre_data_load(batch_train, file_json_train, is_multi_scale, mode_train, num_workers, path_img_train,
                   transform_train):
    dataset_train = CustomCocoDataset(
        file_json=file_json_train,
        path_img=path_img_train,
        mode=mode_train,
        transform=transform_train,
        mode_balance_data=None,
    )
    dataloader_train = torch.utils.data.DataLoader(
        dataset_train,
        batch_size=batch_train,
        num_workers=num_workers,
        shuffle=True,
        pin_memory=True,  # 不使用虚拟内存 GPU要报错
        # drop_last=True,  # 除于batch_size余下的数据
        # collate_fn=lambda x: fun_dataloader(x, cfg),
        collate_fn=CLS4collate_fn(is_multi_scale),
    )
    return dataloader_train



import numpy as np
import cv2
import torch
from PIL import ImageFont
from torchvision import transforms
import matplotlib.pyplot as plt

from tools.datas.f_data_pretreatment4np import f_recover_normalization4ts

'''
plt 常用颜色: 'lightgreen' 'red' 'tan'
'''


def _draw_box4plt(boxes, texts=None, font_size=10, color='red', ax=None, recover_sizewh=None):
    '''

    :param boxes:
    :param texts: 与 boxes 对应的 文本 list
    :param font_size:
    :param color:
    :param ax:  ax = plt.gca()
    :return:
    '''
    if recover_sizewh is not None:
        whwh = np.tile(np.array(recover_sizewh), 2)  # 整体复制 tile
        boxes = boxes * whwh

    if texts is not None:
        try:
            font = ImageFont.truetype('simhei.ttf', font_size, encoding='utf-8')  # 参数1：字体文件路径，参数2：字体大小
        except IOError:
            font = ImageFont.load_default()
        text_w, text_h = font.getsize(texts[0])
        margin = np.ceil(0.05 * text_h)

    for i, box in enumerate(boxes):
        l, t, r, b = box
        w = r - l
        h = b - t

        if ax is None:
            plt.Rectangle((l, t), w, h, color=color, fill=False, linewidth=1)
        else:
            ax.add_patch(plt.Rectangle((l, t), w, h, color=color, fill=False, linewidth=1))

        if texts is not None:
            if ax is None:
                plt.text(l + margin, t - text_h - margin, texts[i],
                         color="r", ha='center', va='center',
                         bbox={'facecolor': 'blue', 'alpha': 0.5, 'pad': 1})
            else:
                ax.text(l + margin, t - text_h - margin, texts[i],
                        color="r", ha='center', va='center',
                        bbox={'facecolor': 'blue', 'alpha': 0.5, 'pad': 1})

        x = l + w / 2
        y = t + h / 2
        plt.scatter(x, y, marker='x', color=color, s=40, label='First')


def f_show_od_ts4plt_v3(img_ts, g_ltrb=None, p_ltrb=None, is_recover_size=False,
                        g_texts=None, is_normal=False):
    assert isinstance(img_ts, torch.Tensor)
    # c,h,w -> h,w,c
    if is_normal:
        img_ts = img_ts.clone()
        img_ts = f_recover_normalization4ts(img_ts)
    img_np_rgb = img_ts.cpu().numpy().astype(np.uint8).transpose((1, 2, 0))
    f_show_od_np4plt_v3(img_np_rgb, g_ltrb=g_ltrb, p_ltrb=p_ltrb,
                        is_recover_size=is_recover_size,
                        g_texts=g_texts)


def f_show_od_np4plt_v3(img_np, g_ltrb=None, p_ltrb=None, other_ltrb=None,
                        is_recover_size=False, g_texts=None):
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np)
    # plt.show()
    ax = plt.gca()

    if is_recover_size:
        recover_sizewh = img_np.shape[:2][::-1]  # npwh
    else:
        recover_sizewh = None

    if p_ltrb is not None:
        plt.title(
            '%s x %s (num_pos = %s)' % (str(img_np.shape[1]), str(img_np.shape[0]), str(len(p_ltrb))))
    else:
        plt.title('%s x %s ' % (str(img_np.shape[1]), str(img_np.shape[0])))

    if other_ltrb is not None:
        _draw_box4plt(other_ltrb, color='tan', ax=ax, recover_sizewh=recover_sizewh)

    if g_ltrb is not None:
        _draw_box4plt(g_ltrb, texts=g_texts, color='lightgreen', ax=ax, recover_sizewh=recover_sizewh)

    if p_ltrb is not None:
        _draw_box4plt(p_ltrb, color='red', ax=ax, recover_sizewh=recover_sizewh)

    plt.show()


# ********************   新方法在上面

def keypoint_painter(images, maps, img_h, img_w, numpy_array=False,
                     phase_gt=False, center_map=None):
    images = images.clone().cpu().data.numpy().transpose([0, 2, 3, 1])
    maps = maps.clone().cpu().data.numpy()
    if center_map is not None:
        center_map = center_map.clone().cpu().data.numpy()
    imgs_tensor = []
    if phase_gt:
        for img, map, c_map in zip(images, maps, center_map):
            img = cv2.resize(img, (img_w, img_h))
            for m in map[:14]:
                h, w = np.unravel_index(m.argmax(), m.shape)
                x = int(w * img_w / m.shape[1])
                y = int(h * img_h / m.shape[0])
                img = cv2.circle(img.copy(), (x, y), radius=1, thickness=2, color=(255, 0, 0))
            h, w = np.unravel_index(c_map.argmax(), c_map.shape)
            x = int(w * img_w / c_map.shape[1])
            y = int(h * img_h / c_map.shape[0])
            img = cv2.circle(img.copy(), (x, y), radius=1, thickness=2, color=(0, 0, 255))
            if numpy_array:
                imgs_tensor.append(img.astype(np.uint8))
            else:
                imgs_tensor.append(transforms.ToTensor()(img))
    else:

        for img, map_6 in zip(images, maps):
            img = cv2.resize(img, (img_w, img_h))
            for step_map in map_6:
                img_copy = img.copy()
                for m in step_map[:14]:
                    h, w = np.unravel_index(m.argmax(), m.shape)
                    x = int(w * img_w / m.shape[1])
                    y = int(h * img_h / m.shape[0])
                    img_copy = cv2.circle(img_copy.copy(), (x, y), radius=1, thickness=2, color=(255, 0, 0))
                if numpy_array:
                    imgs_tensor.append(img_copy.astype(np.uint8))
                else:
                    imgs_tensor.append(transforms.ToTensor()(img_copy))
    return imgs_tensor


def _convert_uint8(img_np):
    if img_np.dtype is not np.uint8:
        img_np_uint8 = img_np.copy()
        img_np_uint8 = img_np_uint8.astype(np.uint8)
    else:
        return img_np
    return img_np_uint8


def f_show_pic_np4plt(img_np):
    '''
    不支持float32
    :param pic:
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np)
    plt.show()


def f_show_pic_np4cv(img_np):
    img_np = _convert_uint8(img_np)
    cv2.imshow('Example', img_np)
    cv2.waitKey(0)


def f_show_kp_np4plt(img_np, boxes_ltrb_input, kps_xy_input, mask_kps, g_ltrb=None,
                     plabels_text=None, p_scores_float=None,
                     is_recover_size=True):
    '''

    :param img_np:
    :param boxes_ltrb_input:
    :param kps_xy_input:
    :param mask_kps:  预测时自动生成一个  torch.ones_like(p_keypoints_f, dtype=torch.bool)
    :param g_ltrb:
    :param plabels_text: [类型名称,]
    :param p_scores_float:[对应分数的list]
    :param is_recover_size:
    :return:
    '''
    img_np = _convert_uint8(img_np)

    assert isinstance(img_np, np.ndarray)
    # import matplotlib.pyplot as plt
    plt.title('%s x %s (num_pos = %s)' % (str(img_np.shape[1]), str(img_np.shape[0]), str(len(boxes_ltrb_input))))
    # plt.imshow(img_np)
    # plt.show()
    whwh = np.tile(np.array(img_np.shape[:2][::-1]), 2)  # 整体复制
    # plt.figure(whwh)  #要报错
    plt.imshow(img_np, alpha=0.7)
    current_axis = plt.gca()
    if g_ltrb is not None:
        if is_recover_size:
            g_ltrb = g_ltrb * whwh
        for box in g_ltrb:
            l, t, r, b = box
            plt_rectangle = plt.Rectangle((l, t), r - l, b - t, color='lightcyan', fill=False, linewidth=3)
            current_axis.add_patch(plt_rectangle)
            # x, y = c[:2]
            # r = 4
            # # 空心
            # # draw.arc((x - r, y - r, x + r, y + r), 0, 360, fill=color)
            # # 实心
            # draw.chord((x - r, y - r, x + r, y + r), 0, 360, fill=_color)

    for i, box_ltrb_input in enumerate(boxes_ltrb_input):
        l, t, r, b = box_ltrb_input
        # ltwh
        current_axis.add_patch(plt.Rectangle((l, t), r - l, b - t, color='green', fill=False, linewidth=2))
        _xys = kps_xy_input[i][mask_kps[i]]

        ''' 就这里不一样 与f_plt_od_np'''
        # 这个是多点连线 加参数后 marker='o' 失效,变成多点连线
        plt.scatter(_xys[::2], _xys[1::2],
                    color='r', s=3, alpha=0.5)
        # plt.plot(_xys[::2], _xys[1::2],
        #          marker='o',
        #          markerfacecolor='red',  # 点颜⾊
        #          markersize=3,  # 点⼤⼩
        #          markeredgecolor='green',  # 点边缘颜⾊
        #          markeredgewidth=2,  # 点边缘宽度
        #          )

        if plabels_text is not None:
            # labels : tensor -> int
            show_text = text = "{} : {:.3f}".format(plabels_text[i], p_scores_float[i])
            current_axis.text(l + 8, t - 10, show_text, size=8, color='white',
                              bbox={'facecolor': 'blue', 'alpha': 0.1})
    plt.show()


def f_show_cpm4input(img_np, kps_xy_input, heatmap_center_input):
    '''

    :param img_np:
    :param kps_xy_input: (14, 2)
    :param heatmap_center_input: (368, 368)
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np, alpha=0.7)
    for xy in kps_xy_input:
        plt.scatter(xy[0], xy[1], color='r', s=5, alpha=0.5)
    plt.imshow(heatmap_center_input, alpha=0.5)
    plt.show()


def f_show_cpm4t_all(img_np, kps_xy_input, heatmap_center_input, heatmap_t, size_wh_input):
    '''
    弄到input进行处理
    :param img_np:
    :param kps_xy_t: (14, 2)
    :param heatmap_center_input: (368, 368,1)
    :param heatmap_t: (46, 46, 14)
    :param size_wh_t: (46, 46, 15)
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np, alpha=1.0)
    for xy in kps_xy_input:
        plt.scatter(xy[0], xy[1], color='r', s=5, alpha=0.5)
    plt.imshow(heatmap_center_input, alpha=0.3)

    h, w, c = heatmap_t.shape
    img_heatmap_z = np.zeros((size_wh_input[1], size_wh_input[0], 1), dtype=np.uint8)
    for i in range(c):
        img_heatmap = heatmap_t[..., i][..., None]
        img_heatmap = cv2.resize(img_heatmap, size_wh_input)  # wh
        # img_heatmap = _convert_uint8(img_heatmap)
        img_heatmap_z = np.maximum(img_heatmap_z, img_heatmap[..., None])

    plt.imshow(img_heatmap_z, alpha=0.3)
    plt.show()


if __name__ == '__main__':
    file_img = r'D:\tb\tb\ai_code\fkeypoint\_test_pic\street.jpg'
    img_np_bgr = cv2.imread(file_img)
    boxes = np.array([
        [1, 1, 100, 100],
        [120, 120, 200, 200],
    ])
    gtexts = ['123', '445566']

    f_show_od_np4plt_v3(img_np_bgr, g_ltrb=boxes, g_texts=gtexts)



import os
import sys


def get_path_root():
    '''os.path.dirname(os.path.abspath(__file__))'''
    debug_vars = dict((a, b) for a, b in os.environ.items() if a.find('IPYTHONENABLE') >= 0)
    # 根据不同场景获取根目录
    if len(debug_vars) > 0:
        """当前为debug运行时"""
        path_root = sys.path[2]
    elif getattr(sys, 'frozen', False):
        """当前为exe运行时"""
        path_root = os.getcwd()
    else:
        """正常执行"""
        path_root = sys.path[1]
    path_root = path_root.replace("\\", "/")  # 替换斜杠
    return path_root



import torch
import torch.nn as nn

from tools.fmodels.f_fun_activation import Mish, SiLU

'''
模型组件
'''


def conv_same(k):
    p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


class AffineChannel(torch.nn.Module):
    def __init__(self, num_features):
        super(AffineChannel, self).__init__()
        self.weight = torch.nn.Parameter(torch.randn(num_features, ))
        self.bias = torch.nn.Parameter(torch.randn(num_features, ))

    def forward(self, x):
        N = x.size(0)
        C = x.size(1)
        H = x.size(2)
        W = x.size(3)
        transpose_x = x.permute(0, 2, 3, 1)
        flatten_x = transpose_x.reshape(N * H * W, C)
        out = flatten_x * self.weight + self.bias
        out = out.reshape(N, H, W, C)
        out = out.permute(0, 3, 1, 2)
        return out


class FConv2d(torch.nn.Module):
    def __init__(self, in_channels, out_channels,
                 k,  # kernel_size
                 s=1,  # stride
                 p=0,  # padding 等于 None 是自动same
                 d=1,  # dilation空洞
                 g=1,  # g groups 一般不动
                 is_bias=True,
                 norm='bn',  # None bn,gn,af
                 act='leaky',  # None relu leaky mish silu identity
                 is_freeze=False,
                 use_dcn=False):
        '''
        有 bn 建议不要 bias
        # 普通conv

        # 降维
        '''
        super(FConv2d, self).__init__()
        self.groups = g
        self.act = act
        self.is_freeze = is_freeze
        self.use_dcn = use_dcn

        if p is None:
            # conv默认为0
            p = conv_same(k)

        # conv
        if use_dcn:
            pass
        else:
            self.conv = torch.nn.Conv2d(in_channels, out_channels, kernel_size=k, stride=s,
                                        padding=p, dilation=d, bias=is_bias)
        # 正则化方式 normalization
        if norm == 'bn':
            self.normalization = torch.nn.BatchNorm2d(out_channels)
        elif norm == 'gn':
            self.normalization = torch.nn.GroupNorm(num_groups=g, num_channels=out_channels)
        elif norm == 'af':
            self.normalization = AffineChannel(out_channels)
        else:
            self.normalization = None

        # act
        if act == 'relu':
            self.act = torch.nn.ReLU(inplace=True)
        elif act == 'leaky':
            self.act = torch.nn.LeakyReLU(0.1, inplace=True)
        elif act == 'mish':
            self.act = Mish()
        elif act == 'silu':
            self.act = SiLU()
        elif act == 'identity':
            self.act = nn.Identity()
        else:
            self.act = None

        self.name_act = act

        if is_freeze:  # 一般不锁定
            self.freeze()

    def freeze(self):
        # 冻结
        self.conv.weight.requires_grad = False
        if self.conv.bias is not None:
            self.conv.bias.requires_grad = False
        if self.bn is not None:
            self.bn.weight.requires_grad = False
            self.bn.bias.requires_grad = False
        if self.gn is not None:
            self.gn.weight.requires_grad = False
            self.gn.bias.requires_grad = False
        if self.act is not None:
            self.act.weight.requires_grad = False
            self.act.bias.requires_grad = False

    def forward(self, x):
        x = self.conv(x)
        if self.normalization is not None:
            x = self.normalization(x)
        if self.act is not None:
            x = self.act(x)
        return x


class DepthwiseConvModule(nn.Module):

    def __init__(self, in_channel, out_channel):
        super().__init__()
        self.depthwise = nn.Conv2d(in_channel, out_channel, 3, padding=1, groups=in_channel, bias=False)
        self.pointwise = nn.Conv2d(in_channel, out_channel, 1, bias=False)
        self.dwnorm = nn.BatchNorm2d(in_channel, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        self.pwnorm = nn.BatchNorm2d(in_channel, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        self.act = nn.LeakyReLU(negative_slope=0.1, inplace=True)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        x = self.dwnorm(x)
        x = self.pwnorm(x)
        x = self.act(x)
        return x



import torch
import numpy as np


def f_cre_grid_cells(size, stride=1, is_center=False,
                     is_flatten=True, is_swap=True, num_repeat=1,
                     device=torch.device('cpu')
                     ):
    '''

    :param size: 决定遍历顺序 第一维fix先固定   第二维b先遍历
    :param stride: 这个为1 为size大小的索引
    :param is_center:  网点是否居中
        off_xy 偏移量 通常为 0~1
    :param is_flatten: 通常这个不要动 因为最后一句拉平 所有无效
    :param is_swap: 是否交换size顺序
    :param num_repeat: 单体复制数 用于批量匹配anc
    :return:
        例 [10, 15] 表示row10(y) col15(x)
            用于遍历矩阵index 采用 is_swap =False 表示先列后行 [[0,0],[0,1]...]
            用于坐标 is_swap =True 表示xy点遍历 [[0,0],[1,0]...]
        [150, 2]
    '''
    fix, b = size  # fix 为先固定  b先遍历
    if is_center:
        off_xy = 0.5
    else:
        off_xy = 0
    x_b_ = (torch.arange(b, dtype=torch.float, device=device) + off_xy) * stride
    y_fix_ = (torch.arange(fix, dtype=torch.float, device=device) + off_xy) * stride
    y_fix, x_b = torch.meshgrid(y_fix_, x_b_)  # 第一个固定
    if is_flatten:
        y_fix = y_fix.flatten()  # torch.Size([10, 15])
        x_b = x_b.flatten()  # torch.Size([10, 15])
    if is_swap:
        stack = torch.stack((x_b, y_fix), dim=-1)
    else:
        stack = torch.stack((y_fix, x_b), dim=-1)

    # 这个直接拉平
    stack = torch.repeat_interleave(stack.view(-1, 2), num_repeat, dim=0)
    return stack
    # return y, x


def guassian_kernel(size_w, size_h, center_x, center_y, sigma):
    gridy, gridx = np.mgrid[0:size_h, 0:size_w]
    D2 = (gridx - center_x) ** 2 + (gridy - center_y) ** 2
    return np.exp(-D2 / 2.0 / sigma / sigma)


if __name__ == '__main__':
    index_rowcol = f_cre_grid_cells([10, 15], stride=1, is_center=False, is_swap=True,
                                    is_flatten=True, num_repeat=3)
    print(index_rowcol)



import torch
from torch import nn
import torch.nn.functional as F

'''
功能层
'''


class Integral(nn.Module):
    '''
    输出最终特图预测值  8 -> 1
    将按 (reg_max+1)*回归个数 的表示,通过期望求出结果
        先将回归值表示成分布概率: softmax
        再离散化: 通过 reg_max 间隙为1离散数组
        求期望: 通过全连接层

    preg = torch.rand(2, 3, 32)
    integral = Integral(7)
    res = integral(preg)
    print(res.shape)
    print(res)
    '''

    def __init__(self, reg_max=16):
        super(Integral, self).__init__()
        self.reg_max = reg_max
        # 7 生成 tensor([0., 1., 2., 3., 4., 5., 6., 7.]) 的区间
        self.register_buffer('project', torch.linspace(0, self.reg_max, self.reg_max + 1))

    def forward(self, x):
        '''
        设数据最大为7 形成[0., 1., 2., 3., 4., 5., 6., 7.] 分个点

        :param x: (2, 3, 32=4*8)
        :return:(6,4) 输出为 拉直的2D
        '''
        # (2,3,32) -> (2*3*4,8)=[24, 8] 形成该层24个点的分布图
        x = F.softmax(x.reshape(-1, self.reg_max + 1), dim=-1)
        # [24, 8] ^^ [8,1] -> [24,1] 通过期望求出每个点的具体值   (一个框有4个点)
        x = F.linear(x, self.project.type_as(x)).reshape(-1, 4)
        return x


if __name__ == '__main__':
    preg = torch.rand(2, 3, 32)
    integral = Integral(7)
    res = integral(preg)
    print(res.shape)
    print(res)



import setuptools

'''
打包上传
python setup.py sdist 
twine upload dist/*
'''
setuptools.setup(
    name="FEADRE_AI",  # Replace with your own username  #自定义封装模块名与文件夹名相同
    version="1.0.1",  # 版本号，下次修改后再提交的话只需要修改当前的版本号就可以了
    author="FEADRE",  # 作者
    author_email="318740003@qq.com",  # 邮箱
    description="My short description for my project.",  # 描述
    long_description='My short description for my project.',  # 描述
    long_description_content_type="text/markdown",  # markdown
    url="https://gitee.com/feadre/feadre-ai.git",  # github地址
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # License
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',  # 支持python版本
)



import cv2
import torch
from PIL import Image
from torchvision import transforms
from torchvision.transforms import functional as F
import numpy as np
import matplotlib.pyplot as plt

'''
PIL九种模型为1，L，P，RGB，RGBA，CMYK，YCbCr，I，F
    “1”为二值图像
    “L”为灰色图像
    “P”为8位彩色图像
    “RGBA”为32位彩色图像 红色、绿色和蓝色 透明
    “CMYK”为32位彩色图像 
        C：Cyan = 青色
        M：Magenta = 品红色
        Y：Yellow = 黄色
        K：Key Plate(blacK) = 定位套版色（黑色）
    “YCbCr”为24位彩色图像
        YCbCr其中Y是指亮度分量
        Cb指蓝色色度分量
        而Cr指红色色度分量
    “I”为32位整型灰色图像
    “F”为32位浮点灰色图像
    
'''


def f转换(file_img):
    # img_np = cv2.imread(file_img) # 这个打开是hwc bgr
    # img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB)

    # -------- tensor2 PIL --------- 两种方式等价
    # plt 支持.convert('RGB') 转换通道
    img_pil = Image.open(file_img)  # 这个打开的就是RGB wh 500*335
    img_tensor = F.to_tensor(img_pil)  # 转换后进行归一化 c,h,w
    transform2tensor = transforms.ToTensor()
    img_tensor = transform2tensor(img_pil)
    img_pil = F.to_pil_image(img_tensor)
    transform2pic = transforms.ToPILImage(mode="RGB")
    img_pil = transform2pic(img_tensor)

    # -------- ts np 正常转--------
    # 直接转需还原归一化 c,h,w -> h,w,c
    img_np_rgb = np.transpose(np.uint8((img_tensor * 255).numpy()), (1, 2, 0))
    img_np_bgr = cv2.cvtColor(img_np_rgb, cv2.COLOR_RGB2BGR)
    img_np_rgb = cv2.cvtColor(img_np_bgr, cv2.COLOR_BGR2RGB)
    # cv2.imshow("img", img_np_bgr)  # 显示
    # key = cv2.waitKey(0)
    # if key == 27:  # 按esc键时，关闭所有窗口
    #     print(key)
    #     cv2.destroyAllWindows()

    # ---------np plt---------
    img_np_rgb = np.array(img_pil)
    img_pil = Image.fromarray(img_np_rgb, mode="RGB")  # 不支持BGR
    img_pil = Image.fromarray(img_np_rgb.astype('uint8')).convert('RGB')
    # img_pil.show()


if __name__ == '__main__':
    file_img = r'D:\tb\tb\ai_code\DL\_test_pic\2007_000042.jpg'  # 500 335
    f转换(file_img)

    pass



from pycocotools.cocoeval import COCOeval
import numpy as np


class FCOCOeval(COCOeval):
    '''
    添加了每一个类的得分
    原装生成运行后自动赋值 coco_eval_obj.stats
    '''

    def __init__(self, cocoGt=None, cocoDt=None, iouType='segm'):
        super().__init__(cocoGt, cocoDt, iouType)

    def summarize(self, catId=None):
        """
        Compute and display summary metrics for evaluation results.
        Note this functin can *only* be applied on the default parameter setting
        """

        def _summarize(ap=1, iouThr=None, areaRng='all', maxDets=100):
            p = self.params
            iStr = ' {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}'
            titleStr = 'Average Precision' if ap == 1 else 'Average Recall'
            typeStr = '(AP)' if ap == 1 else '(AR)'
            iouStr = '{:0.2f}:{:0.2f}'.format(p.iouThrs[0], p.iouThrs[-1]) \
                if iouThr is None else '{:0.2f}'.format(iouThr)

            aind = [i for i, aRng in enumerate(p.areaRngLbl) if aRng == areaRng]
            mind = [i for i, mDet in enumerate(p.maxDets) if mDet == maxDets]

            if ap == 1:
                # dimension of precision: [TxRxKxAxM]
                s = self.eval['precision']
                # IoU
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]

                ''' 这里是添加的代码 判断是否传入catId，如果传入就计算指定类别的指标 '''
                if isinstance(catId, int):
                    s = s[:, :, catId, aind, mind]
                else:
                    s = s[:, :, :, aind, mind]

            else:
                # dimension of recall: [TxKxAxM]
                s = self.eval['recall']
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]

                ''' 这里是添加的代码 判断是否传入catId，如果传入就计算指定类别的指标 '''
                if isinstance(catId, int):
                    s = s[:, catId, aind, mind]
                else:
                    s = s[:, :, aind, mind]

            if len(s[s > -1]) == 0:
                mean_s = -1
            else:
                mean_s = np.mean(s[s > -1])

            ''' 这里是修改 原始只有一个返回值 '''
            # print(iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s))
            # return mean_s
            print_string = iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s)
            return mean_s, print_string

        stats, print_list = [0] * 12, [""] * 12
        stats[0], print_list[0] = _summarize(1)
        stats[1], print_list[1] = _summarize(1, iouThr=.5, maxDets=self.params.maxDets[2])
        stats[2], print_list[2] = _summarize(1, iouThr=.75, maxDets=self.params.maxDets[2])
        stats[3], print_list[3] = _summarize(1, areaRng='small', maxDets=self.params.maxDets[2])
        stats[4], print_list[4] = _summarize(1, areaRng='medium', maxDets=self.params.maxDets[2])
        stats[5], print_list[5] = _summarize(1, areaRng='large', maxDets=self.params.maxDets[2])
        stats[6], print_list[6] = _summarize(0, maxDets=self.params.maxDets[0])
        stats[7], print_list[7] = _summarize(0, maxDets=self.params.maxDets[1])
        stats[8], print_list[8] = _summarize(0, maxDets=self.params.maxDets[2])
        stats[9], print_list[9] = _summarize(0, areaRng='small', maxDets=self.params.maxDets[2])
        stats[10], print_list[10] = _summarize(0, areaRng='medium', maxDets=self.params.maxDets[2])
        stats[11], print_list[11] = _summarize(0, areaRng='large', maxDets=self.params.maxDets[2])

        print_info = "\n".join(print_list)

        if not self.eval:
            raise Exception('Please run accumulate() first')

        self.stats = stats  # 与原装保持一致

        return stats, print_info

    def print_clses(self, clses_name):
        # calculate COCO info for all classes
        # coco_stats, print_coco = self.summarize()

        # calculate voc info for every classes(IoU=0.5)
        voc_map_info_list = []
        for i in range(len(clses_name)):
            stats, _ = self.summarize(catId=i)
            voc_map_info_list.append(" {:15}: {:.3f} {:.3f}".format(clses_name[i], stats[1], stats[7]))

        print_voc = "\n".join(voc_map_info_list)
        print(print_voc)

        # # 将验证结果保存至txt文件中
        # with open("record_mAP.txt", "w") as f:
        #     record_lines = ["COCO results:",
        #                     print_coco,
        #                     "",
        #                     "mAP(IoU=0.5) for each category:",
        #                     print_voc]
        #     f.write("\n".join(record_lines))



import os

import cv2
import torch
from pycocotools.coco import COCO
from torch.utils.data import Dataset
import numpy as np

from tools.GLOBAL_LOG import flog


class CustomCocoDataset(Dataset):
    '''
    v 20210718
    '''

    def __init__(self, file_json, path_img, mode, transform=None, is_debug=False,
                 s_ids_cats=None, nums_cat=None, is_ts_all=True,
                 image_size=None, mode_balance_data=None, name='train', ):
        '''

        :param cfg:
        :param file_json:
        :param path_img:
        :param mode:  bbox segm keypoints caption
        :param transform:
        :param is_debug:
        :param s_ids_cats:  指定类别数
        :param nums_cat:  限制类别的最大数量
        :param is_ts_all:  默认全部转TS
        :param image_size:
        :param mode_balance_data:  数据类别平衡方法
            'max': 4舍五入倍数整体复制  确保类别尽量一致
            'min': 取最少的类型数,多的随机选
            None: 不处理
        :param name: dataset 标注
        '''
        # assert cfg is not None, 'CustomCocoDataset  cfg 不能为空'
        self.image_size = image_size

        self.file_json = file_json
        self.transform = transform
        self.mode = mode
        self.coco_obj = COCO(file_json)

        self.name = name
        self.is_ts_all = is_ts_all

        # f_look_coco_type(self.coco_obj, ids_cats_ustom=None)
        print('创建dataset-----', name)

        if mode_balance_data is not None:
            cats = self.coco_obj.getCatIds()
            num_class = len(cats)
            _t_ids = []
            _t_len = np.zeros(num_class)
            for i, cat in enumerate(cats):
                ids = self.coco_obj.getImgIds(catIds=cat)
                _t_ids.append(ids)
                _t_len[i] = len(ids)

            self.ids_img = []

            if mode_balance_data == 'max':
                num_repeats = np.around((_t_len.max() / _t_len)).astype(np.int)
                for i in range(num_class):
                    self.ids_img.extend(np.tile(np.array(_t_ids[i]), num_repeats[i]).tolist())
            elif mode_balance_data == 'min':
                len_min = _t_len.min().astype(np.int)
                # flog.debug('_t_len = %s', _t_len)
                for i in range(num_class):
                    self.ids_img.extend(np.random.choice(_t_ids[i], len_min).tolist())

        else:
            if s_ids_cats is not None:
                flog.warning('指定coco类型 %s', self.coco_obj.loadCats(s_ids_cats))
                self.s_ids_cats = s_ids_cats
                ids_img = []

                # 限制每类的最大个数
                if nums_cat is None:
                    for idc in zip(s_ids_cats):
                        # 类型对应哪些文件 可能是一张图片多个类型
                        ids_ = self.coco_obj.getImgIds(catIds=idc)
                        ids_img += ids_
                else:
                    # 限制每类的最大个数
                    for idc, num_cat in zip(s_ids_cats, nums_cat):
                        # 类型对应哪些文件 可能是一张图片多个类型
                        ids_ = self.coco_obj.getImgIds(catIds=idc)[:num_cat]
                        # ids_ = self.coco.getImgIds(catIds=idc)[:1000]
                        ids_img += ids_
                        # print(ids_)  # 这个只支持单个元素

                self.ids_img = list(set(ids_img))  # 去重
            else:
                self.ids_img = self.coco_obj.getImgIds()  # 所有图片的id 画图数量

        #  创建 coco 类别映射
        self._init_load_classes()  # 除了coco数据集,其它不管

        self.is_debug = is_debug
        self.path_img = path_img
        if not os.path.exists(path_img):
            raise Exception('coco path_img 路径不存在', path_img)

    def __len__(self):
        return len(self.ids_img)

    def open_img_tar(self, id_img):
        img = self.load_image(id_img)

        # bboxs, labels,keypoints
        tars_ = self.load_anns(id_img, img_wh=img.shape[:2][::-1])
        if tars_ is None:  # 没有标注返回空
            return None

        # 动态构造target
        target = {}
        l_ = ['boxes', 'labels', 'keypoints', 'kps_mask']
        target['image_id'] = id_img
        target['size'] = np.array(img.shape[:2][::-1])  # (w,h)

        # 根据标注模式 及字段自动添加 target['boxes', 'labels', 'keypoints']
        for i, tar in enumerate(tars_):
            target[l_[i]] = tar
        return img, target

    def __getitem__(self, index):
        '''

        :param index:
        :return: tensor or np.array 根据 out: 默认ts or other is np
            img: h,w,c
            target:
            coco原装是 ltwh
            dict{
                image_id: int,
                bboxs: ts n4 原图 ltwh -> ltrb
                labels: ts n,
                keypoints: ts n,10
                size: wh
            }
        '''
        # 这里生成的是原图尺寸的 target 和img_np_uint8 (375, 500, 3)
        id_img = self.ids_img[index]
        res = self.open_img_tar(id_img)

        if res is None:
            print('这个图片没有标注信息 id为 %s ,继续下一个', id_img)
            return self.__getitem__(index + 1)

        img, target = res

        _text_base = '!!! 数据有问题 %s  %s %s %s '
        assert len(target['boxes']) == len(target['labels']), \
            _text_base % ('transform前', target, len(target['boxes']), len(target['labels']))

        if target['boxes'].shape[0] == 0:
            flog.warning('数据有问题 重新加载 %s', id_img)
            return self.__getitem__(index + 1)

        # 以上img 确定是 np格式(transform出来一般是ts); target 全部转np
        if self.transform is not None:
            img, target = self.transform(img, target)
        # 这里img输出 ts_3d
        if self.is_ts_all:
            target['boxes'] = torch.tensor(target['boxes'], dtype=torch.float)
            target['labels'] = torch.tensor(target['labels'], dtype=torch.int64)
            target['size'] = torch.tensor(target['size'], dtype=torch.float)  # file尺寸

        assert len(target['boxes']) == len(target['labels']), \
            _text_base % ('transform后', target, len(target['boxes']), len(target['labels']))
        # 每个图片对应的target数量是不一致的 所以需要用target封装
        return img, target

    def load_image(self, id_img):
        '''

        :param id_img:
        :return:
        '''
        image_info = self.coco_obj.loadImgs(id_img)[0]
        file_img = os.path.join(self.path_img, image_info['file_name'])
        if not os.path.exists(file_img):
            raise Exception('file_img 加载图片路径错误', file_img)

        img = cv2.imread(file_img)
        return img

    def load_anns(self, id_img, img_wh):
        '''
        ltwh --> ltrb
        :param id_img:
        :return:
            bboxs: np(num_anns, 4)
            labels: np(num_anns)
        '''
        # annotation_ids = self.coco.getAnnIds(self.image_ids[index], iscrowd=False)
        annotation_ids = self.coco_obj.getAnnIds(id_img)  # ann的id
        # anns is num_anns x 4, (x1, x2, y1, y2)
        bboxs_np = np.zeros((0, 4), dtype=np.float32)  # np创建 空数组 高级
        labels = []
        if len(annotation_ids) == 0:
            return None

        coco_anns = self.coco_obj.loadAnns(annotation_ids)
        for ann in coco_anns:
            x, y, box_w, box_h = ann['bbox']  # ltwh
            # 修正超范围的框  得 ltrb
            x1 = max(0, x)  # 修正lt最小为0 左上必须在图中
            y1 = max(0, y)
            x2 = min(img_wh[0] - 1, x1 + max(0, box_w - 1))  # 右下必须在图中
            y2 = min(img_wh[1] - 1, y1 + max(0, box_h - 1))
            ''' bbox校验 '''
            if ann['area'] > 0 and x2 > x1 and y2 >= y1:
                bbox = np.array([[x1, y1, x2, y2]], dtype=np.float32)
                # bbox[0, :4] = ann['bbox']
                # ann['bbox'] = [x1, y1, x2, y2]  # 这样写回有BUG 共享内存会修改
            else:
                flog.error('标记框有问题 %s 跳过', ann)
                continue

            # 全部通过后添加
            bboxs_np = np.append(bboxs_np, bbox, axis=0)
            labels.append(self.classes_coco2train[ann['category_id']])

        # bboxs = ltwh2ltrb(bboxs) # 前面 已转
        if bboxs_np.shape[0] == 0:
            flog.error('这图标注 不存在 %s', coco_anns)
            return None
            # raise Exception('这图标注 不存在 %s', coco_anns)

        # 这里转tensor
        if self.mode == 'bbox':
            return [bboxs_np, labels]
        elif self.mode == 'keypoints':
            pass

    def _init_load_classes(self):
        '''
        self.classes_ids :  {'Parade': 1}
        self.ids_classes :  {1: 'Parade'}
        self.ids_new_old {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10, 11: 11, 12: 12, 13: 13, 14: 14, 15: 15, 16: 16, 17: 17, 18: 18, 19: 19, 20: 20}
        self.ids_old_new {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10, 11: 11, 12: 12, 13: 13, 14: 14, 15: 15, 16: 16, 17: 17, 18: 18, 19: 19, 20: 20}
        :return:
        '''
        # [{'id': 1, 'name': 'aeroplane'}, {'id': 2, 'name': 'bicycle'}, {'id': 3, 'name': 'bird'}, {'id': 4, 'name': 'boat'}, {'id': 5, 'name': 'bottle'}, {'id': 6, 'name': 'bus'}, {'id': 7, 'name': 'car'}, {'id': 8, 'name': 'cat'}, {'id': 9, 'name': 'chair'}, {'id': 10, 'name': 'cow'}, {'id': 11, 'name': 'diningtable'}, {'id': 12, 'name': 'dog'}, {'id': 13, 'name': 'horse'}, {'id': 14, 'name': 'motorbike'}, {'id': 15, 'name': 'person'}, {'id': 16, 'name': 'pottedplant'}, {'id': 17, 'name': 'sheep'}, {'id': 18, 'name': 'sofa'}, {'id': 19, 'name': 'train'}, {'id': 20, 'name': 'tvmonitor'}]
        categories = self.coco_obj.loadCats(self.coco_obj.getCatIds())
        categories.sort(key=lambda x: x['id'])  # 按id升序 [{'id': 1, 'name': 'Parade'}]

        # coco ids is not from 1, and not continue ,make a new index from 0 to 79, continuely
        # 重建index 从1-80
        # classes_ids:   {names:      new_index}
        # coco_ids:  {new_index:  coco_index}
        # coco_ids_inverse: {coco_index: new_index}

        self.classes_ids, self.classes_train2coco, self.classes_coco2train = {}, {}, {}
        self.ids_classes = {}
        # 解决中间有断格的情况
        for i, c in enumerate(categories, start=1):  # 修正从1开始
            self.classes_train2coco[i] = c['id']  # 验证时用这个
            self.classes_coco2train[c['id']] = i
            self.classes_ids[c['name']] = i  # 这个是新索引 {'Parade': 0}
            self.ids_classes[i] = c['name']
        pass


if __name__ == '__main__':
    class cfg:
        pass


    path_root = 'M:/AI/datas/VOC2012'
    path_img = os.path.join(path_root, 'train/JPEGImages')
    file_json = os.path.join(path_root, 'coco/annotations/instances_train_17125.json')
    mode = 'bbox'  # bbox segm keypoints caption
    transform = None

    dataset = CustomCocoDataset(
        cfg=cfg,
        file_json=file_json,
        path_img=path_img,
        mode=mode,
        transform=transform,
        mode_balance_data=None,
    )

    print(dataset[0])
    print('len(dataset)', len(dataset))
